/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.io.Serializable;

import de.jtem.mfc.field.Complex;
import de.jtem.mfc.geometry.ComplexProjective1;
import de.jtem.mfc.geometry.HermitianCircle;
import de.jtem.moebiusViewer.util.MoebiusShapeStack;
import de.jtem.moebiusViewer.util.MoebiusTransformStack;

public abstract class MoebiusGraphics implements Serializable, Cloneable {
    private static final long serialVersionUID = 1;

    protected Attributes        attributes;
    protected Dimension         size = new Dimension();

    protected MoebiusTransform  transform;

    MoebiusTransformStack  transformStack = new MoebiusTransformStack ();

    MoebiusShapeStack      shapeStack     = new MoebiusShapeStack ();

    MoebiusTransform[]     contextStack   = new MoebiusTransform[1];
    int                    contextStackSize;

    /* Moebius geometric primitives */

    public abstract void point (ComplexProjective1 z);
    public abstract void circle (HermitianCircle c);
    public abstract void arc (ComplexProjective1 z1, ComplexProjective1 z2, ComplexProjective1 z3);
    public abstract void lineSeg (ComplexProjective1 z1, ComplexProjective1 z2);
    public abstract void text (ComplexProjective1 z, String s );

    public MoebiusGraphics() {
	this (new Attributes ());
    }

    public MoebiusGraphics( Attributes v) {
	attributes = v;

	transform = contextStack[0]  = new MoebiusTransform();

	contextStackSize = 1;
    }

    public Attributes getAttributes () {
	return attributes;
    }

    public void setAttributes(Attributes v) {
	if (attributes == v) {
	    return;
	}

	if (attributes.graphics != null) {
	    v.setGraphics (attributes.graphics);
	}

	attributes = v;
    }

    public Dimension getSize() {
	return size;
    }

    public void setSize (Dimension d){
        size = d;
    }

    public Graphics getGraphics () {
	return attributes.graphics;
    }

    public void setGraphics (Graphics v) {
	attributes.setGraphics (v);
    }

    public Object clone () {
	try {
	    return super.clone ();
	}
	catch (java.lang.CloneNotSupportedException e1) {
	    return null;
	}
    }

    public void clear() {
	shapeStack.clear();
	transformStack.clear();

	transform = contextStack[0];

	contextStackSize = 1;
    }

    public MoebiusTransform popTransform () {

	if( contextStackSize == 1 )
	    throw new java.util.EmptyStackException();

	transform = contextStack[ --contextStackSize - 1 ];

	return transformStack.pop ();
    }


    final private void assureCapacityOfContextStack( int minimalCapacity ) {

	if( contextStack.length >= minimalCapacity )
	    return;

	final MoebiusTransform [] newContextStack = new MoebiusTransform[ minimalCapacity ];

	System.arraycopy( contextStack, 0, newContextStack, 0, contextStack.length );

	for( int i = contextStack.length; i < newContextStack.length; i++ )
	    newContextStack[i] = new MoebiusTransform();

	contextStack = newContextStack;
    }

    private void updateContextStack() {

	assureCapacityOfContextStack( transformStack.size() + 1 );

	contextStackSize = transformStack.size() + 1;

	contextStack[1].assign( transformStack.elementAt( 0 ) );

	for( int i=1; i<transformStack.size(); i++ ) {
	    contextStack[i+1].assignTimes( contextStack[ i ], transformStack.elementAt( i ) );
	}

	transform = contextStack[ contextStackSize - 1 ];
    }

    public MoebiusTransform push (MoebiusTransform T) {

	transformStack.push( T );

	assureCapacityOfContextStack( contextStack.length + 1 );

	contextStack[contextStackSize++].assignTimes( transform, T );

	transform = contextStack[ contextStackSize - 1 ];

	return T;
    }

    final public MoebiusTransform getTransform () {
	return transform;
    }

    final public MoebiusTransformStack getTransformStack() {
	return transformStack;
    }

    public void setTransformStack( MoebiusTransformStack v) {

	transformStack = new MoebiusTransformStack( v );

	updateContextStack();
    }

    public MoebiusShapeStack getShapeStack () {
	return shapeStack;
    }

    public MoebiusShape push (MoebiusShape aShape) {
	shapeStack.push (aShape);

	return aShape;
    }

    public MoebiusShape popShape () {
	return  shapeStack.pop ();
    }

    public int getHorizontalTextLayout () {
	return attributes.horizontalTextLayout;
    }

    public void setHorizontalTextLayout (int aNum) {
	attributes.setHorizontalTextLayout (aNum);
    }

    public int getVerticalTextLayout () {
	return attributes.verticalTextLayout;
    }

    public void setVerticalTextLayout (int aNum) {
	attributes.setVerticalTextLayout (aNum);
    }

    public Font getFont () {
	return attributes.font;
    }

    public void setFont (Font f) {
	attributes.setFont (f);
    }

    public Color getColor () {
	return attributes.color;
    }

    public void setColor (Color v) {
	attributes.setColor (v);
    }

    public Color getPointColor () {
	return attributes.pointColor;
    }

    public void setPointColor (Color v) {
	attributes.setPointColor (v);
    }

    public int getPointRadius () {
	return attributes.pointRadius;
    }

    public void setPointRadius (int v) {
	attributes.setPointRadius (v);
    }

    public double getLineWidth () {
      return attributes.lineWidth;
    }

    public void setLineWidth (double v) {
	attributes.setLineWidth(v);
    }

    public int getPointOutline () {
	return attributes.pointOutline;
    }

    public void setPointOutline (int v) {
	attributes.setPointOutline(v);
    }

    public boolean getFilled () {
	return attributes.filled;
    }

    public void setFilled (boolean v) {
	attributes.setFilled (v);
    }

    public boolean getUsePointColor () {
	return attributes.usePointColor;
    }

    public void setUsePointColor (boolean v) {
	attributes.setUsePointColor (v);
    }

    private HermitianCircle       tc = new HermitianCircle();
    private ComplexProjective1    tz = new ComplexProjective1();
    private ComplexProjective1    tw = new ComplexProjective1();
    private ComplexProjective1    tu = new ComplexProjective1();
    private Complex               z = new Complex();
    private Complex               w = new Complex();
    private Complex               u = new Complex();
    private Complex               t = new Complex();

    public void point( double x, double y) {
	tz.assign(x, y, 1.0, 0.0);
	point(tz);
    }

    public void point( double aRe, double aIm, double bRe, double bIm){
	tz.assign(aRe, aIm, bRe, bIm);
	point(tz);
    }

    public void text( double x, double y, String str ) {
	tz.assign(x, y, 1.0, 0.0);
	text(tz, str );
    }

    public void text( double aRe, double aIm, double bRe, double bIm, String str ){
	tz.assign(aRe, aIm, bRe, bIm);
	text( tz, str );
    }

    public void circle( double x, double y, double r) {
	tc.assignEuclidean(x, y, r);
	circle(tc);
    }

    public void arc( double x1, double y1, double x2, double y2, double x3, double y3){
	tz.assign(x1, y1, 1.0, 0.0);
	tu.assign(x2, y2, 1.0, 0.0);
	tw.assign(x3, y3, 1.0, 0.0);

	arc(tz, tu, tw);
    }

    public void line( double x1, double y1, double x2, double y2) {
	tz.assign(x1, y1, 1.0, 0.0);
	tu.assign(x2, y2, 1.0, 0.0);
	tw.assign(1.0, 0.0, 0.0, 0.0);
	tc.assignThrough(tz, tu, tw);
	circle(tc);
    }

    public void line( double aR1, double aI1, double bR1, double bI1,
		      double aR2, double aI2, double bR2, double bI2){
        tz.assign(aR1, aI1, bR1, bI1);
        tu.assign(aR2, aI2, bR2, bI2);
	tw.assign(1.0, 0.0, 0.0, 0.0);
	tc.assignThrough(tz, tu, tw);
	circle(tc);
    }

    public void lineSeg (double x1, double y1, double x2, double y2) {
	tz.assign(x1, y1, 1.0, 0.0);
	tw.assign(x2, y2, 1.0, 0.0);

	lineSeg(tz, tw);
    }

    public void lineSeg (double aR1, double aI1, double bR1, double bI1,
                         double aR2, double aI2, double bR2, double bI2){
        tz.assign(aR1, aI1, bR1, bI1);
        tw.assign(aR2, aI2, bR2, bI2);

        lineSeg(tz, tw);
    }

}

