/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.moebius;

import de.jtem.mfc.field.Complex;
import de.jtem.mfc.geometry.ComplexProjective1;
import de.jtem.mfc.geometry.HermitianCircle;
import de.jtem.moebiusViewer.AbstractDrawer;
import de.jtem.moebiusViewer.AbstractPSWriter;
import de.jtem.moebiusViewer.Attributes;
import de.jtem.moebiusViewer.MoebiusGraphicsWithPrimitives;

public class MoebiusDrawer extends AbstractDrawer {

    private static final long     serialVersionUID = 1L;

    public AbstractPSWriter getPSWriter() {
	return new MoebiusPSWriter( attributes );
    }

    public void point( ComplexProjective1 aZ ) {
	MoebiusDrawer.point( this, aZ );
    }

    public void circle ( HermitianCircle aC){
	MoebiusDrawer.circle( this, aC );
    }

    public void lineSeg( ComplexProjective1 aZ, ComplexProjective1 aW) {
	MoebiusDrawer.lineSeg( this, aZ, aW );
    }

    public void arc( ComplexProjective1 z1, ComplexProjective1 z2, ComplexProjective1 z3){
	MoebiusDrawer.arc( this, z1, z2, z3 );
    }

    public void text( ComplexProjective1 cp, String s) {
	MoebiusDrawer.text( this, cp, s );
    }

    final static HermitianCircle       tc = new HermitianCircle();

    final static ComplexProjective1    tz = new ComplexProjective1();
    final static ComplexProjective1    tw = new ComplexProjective1();
    final static ComplexProjective1    tu = new ComplexProjective1();

    final static Complex               z  = new Complex();
    final static Complex               w  = new Complex();
    final static Complex               u  = new Complex();
    final static Complex               t  = new Complex();

    static void point( MoebiusGraphicsWithPrimitives graphics, ComplexProjective1 aZ ){

        tz.assignTimes( graphics.getTransform(), aZ );
        tz.getB(z);

	Attributes attributes = graphics.getAttributes();

        if (!z.equals(0.,0.)){
            tz.projectTo(z);

	    graphics.fillColor( attributes.color );

            graphics.fillCircle( z.re, z.im, attributes.pointRadius );

            if( attributes.pointOutline > 0 ) {

		graphics.fillColor( attributes.pointColor );

                graphics.fillCircle
		    ( z.re, z.im,  Math.max( 0, attributes.pointRadius - attributes.pointOutline ) );
            }
        }
    }

    static void circle( MoebiusGraphicsWithPrimitives graphics, HermitianCircle aC ){
        tc.assign(aC);
	tc.apply( graphics.getTransform() );
        tc.euclideanCenter(z);

	Attributes attributes = graphics.getAttributes();

        double radius   = Math.abs(tc.euclideanRadius());
        double diameter = 2*radius;
	graphics.drawColor( attributes.color );
        if (z.isInfinite() || radius > 100000){

            double d = tc.getDRe();
            int gw = graphics.getSize().width;
            int gh = graphics.getSize().height;
            double t = (double)(gh*gh + gw*gw);
            tc.getB(u);
            double n1 = -2* u.re;
            double n2 = -2* u.im;
            double n = Math.sqrt(n1*n1 + n2*n2);
            d  = d/n;
            n1 = n1/n;
            n2 = n2/n;

            graphics.drawLine( (d*n1+t*n2), (d*n2-t*n1),
			       (d*n1-t*n2), (d*n2+t*n1));
        } else {
            if( attributes.filled ) {
                graphics.fillCircle( z.re, z.im, radius );
            } else {
                graphics.drawCircle( z.re, z.im, radius );
            }
        }
    }

    static void lineSeg( MoebiusGraphicsWithPrimitives graphics,
			 ComplexProjective1 aZ, ComplexProjective1 aW) {

        tz.assign(aZ);
        tz.assignTimes( graphics.getTransform(), tz);
        tz.projectTo(z);
        tw.assign(aW);
        tw.assignTimes(graphics.getTransform(), tw);
        tw.projectTo(w);

	Attributes attributes = graphics.getAttributes();

	graphics.drawColor( attributes.color );

        double width = attributes.lineWidth;

        if (!z.isInfinite() && !w.isInfinite()){
            if ((width) == 1 || attributes.getJava2()) {
                graphics.drawLine (z.re, z.im,
                          w.re, w.im);
            } else {

                // thick lines in Java 1

                final double x1t;
                final double x2t;
                final double y1t;
                final double y2t;
                final double xDiff;
                final double yDiff;
                final double norm;
                final double xNormal;
                final double yNormal;
                x1t = z.re;
                y1t = z.im;
                x2t = w.re;
                y2t = w.im;

                xDiff = x2t - x1t;
                yDiff = y2t - y1t;
                norm = Math.sqrt(xDiff * xDiff + yDiff * yDiff);
                yNormal = xDiff / norm;
                xNormal = -yDiff / norm;
                double [] x = new double[4];
                double [] y = new double[4];
                double halfWidth = 0.5*width;

                x[0] = x1t  - halfWidth * xNormal;
                y[0] = y1t  - halfWidth * yNormal;
                x[1] = x2t  - halfWidth * xNormal;
                y[1] = y2t  - halfWidth * yNormal;
                x[2] = x[1] +     width * xNormal;
                y[2] = y[1] +     width * yNormal;
                x[3] = x[0] +     width * xNormal;
                y[3] = y[0] +     width * yNormal;

                graphics.fillPolygon(x, y, 4);
            }
        }
    }

    static void arc( MoebiusGraphicsWithPrimitives graphics,
		     ComplexProjective1 z1, ComplexProjective1 z2, ComplexProjective1 z3){

        tc.assignThrough(z1, z2, z3);
        tc.apply(graphics.getTransform());
        tc.euclideanCenter(u);
        double c1 = u.re;
        double c2 = u.im;

        tz.assign(z1);
        tz.assignTimes(graphics.getTransform(), tz);
        tz.projectTo(z);

        tu.assign(z2);
        tu.assignTimes(graphics.getTransform(), tu);
        tu.projectTo(u);

        tw.assign(z3);
        tw.assignTimes(graphics.getTransform(), tw);
        tw.projectTo(w);

        u.assignMinus(z);
        t.assignMinus(w,z);
        t.assignTimes(0,1);
        t.assignConjugate();
        t.assignTimes(u);
        u.assignPlus(z);

        if( t.re  < 0){
          t.assign(z);
          z.assign(w);
          w.assign(t);
        }

	Attributes attributes = graphics.getAttributes();

	graphics.drawColor( attributes.color );

        if (u.isInfinite() || Math.abs(tc.euclideanRadius()) > 100000){


            if (((u.re > z.re)&&(u.re < w.re))||((u.re < z.re)&&(u.re > w.re))||
                ((u.im > z.im)&&(u.im < w.im))||((u.im < z.im)&&(u.im > w.im))){
                graphics.drawLine (z.re, z.im,
                          w.re, w.im);
                }
            else{
                double d = tc.getDRe();
                int gw = graphics.getSize().width;
                int gh = graphics.getSize().height;
                double l = (double)(gh*gh + gw*gw);
                tc.getB(t);
                double n1 = -2* t.re;
                double n2 = -2* t.im;
                double n = Math.sqrt(n1*n1 + n2*n2);
                d  = d/n;
                n1 = n1/n;
                n2 = n2/n;

                if ( d*n1+l*n2 < d*n1-l*n2 || (d*n1+l*n2 == d*n1-l*n2 && d*n1+l*n2 < d*n2+l*n1)){
                    u.assign(d*n1+l*n2, d*n2-l*n1);
                    t.assign(d*n1-l*n2, d*n2+l*n1);
                }
                else{
                    t.assign(d*n1+l*n2, d*n2-l*n1);
                    u.assign(d*n1-l*n2, d*n2+l*n1);
                }

                if(w.re < z.re || (u.re == t.re && w.im < z.im)){
                    graphics.drawLine(u.re, u.im, w.re, w.im);
                    graphics.drawLine(t.re, t.im, z.re, z.im);
                }
                else{
                    graphics.drawLine(u.re, u.im, z.re, z.im);
                    graphics.drawLine(t.re, t.im, w.re, w.im);
                }
            }
        }
        else{

            double radius  = Math.abs(tc.euclideanRadius());
            double alpha =   Math.toDegrees( Math.atan2(c2-z.im, z.re-c1));
            double beta  = ( Math.toDegrees( Math.atan2(c2-w.im, w.re-c1))) - alpha;

            if(beta < 0){
                beta += 360;
            }

            if(Math.abs(beta) < 1.5){
                graphics.drawLine (z.re, z.im, w.re, w.im);
            }
            else{
                if (beta > 0){
                    alpha += 1;
                    beta -= 1.5;
                }
                else{
                    alpha -= 1;
                    beta += 1.5;
                }

                graphics.drawArc(c1, c2, radius, -alpha, -beta );

                double drawX = (c1 + radius * Math.cos(Math.toRadians( alpha )));
                double drawY = (c2 - radius * Math.sin(Math.toRadians( alpha )));
                graphics.drawLine (z.re, z.im, drawX, drawY);
                drawX = (c1 + radius * Math.cos(Math.toRadians( alpha + beta )));
                drawY = (c2 - radius * Math.sin(Math.toRadians( alpha + beta )));
                graphics.drawLine (drawX, drawY, w.re, w.im);

            }
        }
    }


    static void text(  MoebiusGraphicsWithPrimitives graphics, ComplexProjective1 cp, String s) {
        tu.assignTimes( graphics.getTransform(), cp);
	tu.projectTo(u);

        if (u.isInfinite())
            return;

          graphics.drawColor( graphics.getColor() );
	graphics.drawString( s, u.re, u.im );

	/*getGraphics ().drawLine( u.re, u.im,
				 u.re - fontMetrics.stringWidth (s), u.im );
	getGraphics ().drawLine( u.re, u.im,
	u.re, u.im - fontMetrics.getAscent() );  */

    }

}



