/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import java.awt.Color;

import de.jtem.mfc.field.Complex;
import de.jtem.mfc.geometry.ComplexProjective1;
import de.jtem.moebiusViewer.Attributes;
import de.jtem.moebiusViewer.MoebiusGraphics;

public abstract class AbstractPointSet extends AbstractShape {
    private static final long     serialVersionUID = 1L;
    protected int numOfPoints;

    protected double [] aRe;
    protected double [] aIm;
    protected double [] bRe;
    protected double [] bIm;

    protected Color [] pointColor = new Color[0];

    protected Color defaultPointColor = Color.black;

    protected boolean labelPointsWithIndex = false;
    protected boolean consistent = true;

    protected AbstractPointSet (){
        this(0);
    }

    protected AbstractPointSet ( int n) {
        setNumOfPoints( n, true );
    }
    
    protected AbstractPointSet( double [] coords ) {
	assignCoords (coords);
    }

    protected AbstractPointSet( Complex [] coords ) {
	this( coords.length );
	assignCoords ( coords );
    }

    protected AbstractPointSet( ComplexProjective1 [] coords ) {
	this( coords.length );
	assignCoords ( coords );
    }

    public double [] aRe() {
        return aRe;
    }

    public double [] aIm() {
        return aIm;
    }

    public double [] bRe() {
        return bRe;
    }

    public double [] bIm() {
        return bIm;
    }

    protected boolean isLabelPointsWithIndex() {
	return labelPointsWithIndex;
    }

    protected boolean getLabelPointsWithIndex() {
	return labelPointsWithIndex;
    }

    protected void setLabelPointsWithIndex( boolean aBoolean ) {
        if (labelPointsWithIndex == aBoolean)
            return;

	labelPointsWithIndex = aBoolean;
        firePropertyChange("labelPointsWithIndex");
    }

    protected boolean isConsistent() {
	return consistent;
    }

    protected void setConsistent( boolean aBoolean ) {
        if (consistent == aBoolean)
            return;

	consistent = aBoolean;
        firePropertyChange("consistent");
    }

    protected int getNumOfPoints () {
        
	return numOfPoints;
    } 

    protected void setNumOfPoints (int newNumOfPoints) {

	if( newNumOfPoints == numOfPoints )
	    return;
	setNumOfPoints( newNumOfPoints, true );
	firePropertyChange( "numOfPoints" );
    }

    /** this must not fire property change. */
    protected void setNumOfPoints( int newNumOfPoints, boolean assignDefault ) {

        if( numOfPoints == newNumOfPoints )
            return;

        final int minSize = newNumOfPoints < numOfPoints 
            ? newNumOfPoints : numOfPoints;

        double [] newARe = new double[ newNumOfPoints ];
        double [] newAIm = new double[ newNumOfPoints ];
        double [] newBRe = new double[ newNumOfPoints ];
        double [] newBIm = new double[ newNumOfPoints ];

        if( aRe != null ) {
            System.arraycopy( aRe, 0, newARe, 0, minSize );
            System.arraycopy( aIm, 0, newAIm, 0, minSize );
            System.arraycopy( bRe, 0, newBRe, 0, minSize );
            System.arraycopy( bIm, 0, newBIm, 0, minSize );
        }

        aRe = newARe;
        aIm = newAIm;
        bRe = newBRe;
        bIm = newBIm;

        for( int i = minSize; i<newNumOfPoints; i++ ){
            bRe[i] = 1;
            if (assignDefault)
                aRe[i] = i;
        }
	numOfPoints = newNumOfPoints;
        updatePointColor();
    }

    protected void updatePointColor() {
    
        if( pointColor.length == numOfPoints)
            return;

	if( pointColor == null )
	    return;
	
        Color[] newPointColor = new Color[numOfPoints];
	    
        System.arraycopy( pointColor, 0, newPointColor, 0,
                          Math.min( pointColor.length, newPointColor.length ) );
	
        pointColor = newPointColor;
    } 

    /* methods for ComplexProjective coords */   

    protected void setProjectiveCoords (int index, double aRe1, double aIm1, 
                                         double bRe1, double bIm1){
        if (ComplexProjective1.equals(aRe[index], aIm[index], 
                                      bRe[index], bIm[index], 
                                      aRe1, aIm1, bRe1, bIm1))
            return;

        aRe[index] = aRe1;
        aIm[index] = aIm1;
        bRe[index] = bRe1;
        bIm[index] = bIm1;
        //fireMoebiusShapeChange( MoebiusShapeChangeEvent.COORDS_CHANGED);
	firePropertyChange( "coords" );
    } 

    protected void setProjectiveCoords( int index, final ComplexProjective1 z ) {
    
        setProjectiveCoords(index, z.aRe, z.aIm, z.bRe, z.bIm); 
    } 
       
  protected void setProjectiveCoords( int index, final Complex z, final Complex u ) {
    
    setProjectiveCoords(index, z.re, z.im, u.re, u.im); 
  }  

  
  protected void getProjectiveCoords( ComplexProjective1 [] coords ) {

    if( coords == null || coords.length != numOfPoints )
      throw new IllegalArgumentException( "invalid coords array" );
    
    
    for( int i=0; i<numOfPoints; i++ ) {
      
      if( coords[i] == null ) {
	    coords[i] = new ComplexProjective1( aRe[i], aIm[i], bRe[i], bIm[i] );
      } else {
	coords[i].assign( aRe[i], aIm[i], bRe[i], bIm[i] );
      }
    }
  }
    
  protected ComplexProjective1[] getProjectiveCoordsAsArray1d() {
    
    final ComplexProjective1[] coords = new ComplexProjective1[ numOfPoints ];

    for( int i=0; i<numOfPoints; i++ ) {
      coords[i] = new ComplexProjective1( aRe[i], aIm[i], bRe[i], bIm[i] );
    }
    
    return coords;
  }

  protected void getCoords( Complex [] coords ) {
    
    if( coords == null || coords.length != numOfPoints )
      throw new IllegalArgumentException( "invalid coords array" );
    
    
    for( int i=0; i<numOfPoints; i++ ) {
      
      if( coords[i] == null ) {
	coords[i] = new Complex( aRe[i], aIm[i] );
      } else {
	coords[i].assign( aRe[i], aIm[i] );
      }

      coords[i].assignDivide( bRe[i], bIm[i] );
    }
  }
    
    protected Complex[] getCoordsAsArray1d() {

       	final Complex[] coords = new Complex[ numOfPoints ];

	for( int i=0; i<numOfPoints; i++ ) {
	  Complex c = new Complex( aRe[i], aIm[i] );
	  c.assignDivide( bRe[i], bIm[i] );
	  coords[i] = c;
	}
	return coords;
    }

 
     /** assign all coords at once */
     protected void assignCoords( ComplexProjective1 [] coords ) {

         boolean propertyHasChanged = coords.length != numOfPoints;

         setNumOfPoints(coords.length, false);

         for( int i=0; i<numOfPoints; i++ ) {
            
             final ComplexProjective1 z = coords[i];
            
             propertyHasChanged = propertyHasChanged || !ComplexProjective1.equals
                 ( aRe[i], aIm[i], bRe[i], bIm[i], z.aRe, z.aIm, z.bRe, z.bIm );

             aRe[i] = z.aRe;
             aIm[i] = z.aIm;
             bRe[i] = z.bRe;
             bIm[i] = z.bIm;
         }

	updatePointColor();

        if( propertyHasChanged )
            firePropertyChange( "coords" );
    } 


    protected void getProjectiveCoords(int index, ComplexProjective1 coords){
        coords.assign(aRe[index], aIm[index], bRe[index], bIm[index]);
    }

    protected ComplexProjective1 getProjectiveCoords( int index ) {
        final ComplexProjective1 z = new ComplexProjective1( aRe[index], aIm[index], bRe[index], bIm[index] );
	return z;
    }

  protected Complex getCoords( int index ) {
    Complex c = new Complex( aRe[index], aIm[index] );
    c.assignDivide(  bRe[index], bIm[index] );
    return c;
  }


    /* methods for Field.Complex Coords */
    
    protected void setCoords( int index, final Complex z ) {
        setCoords( index, z.re, z.im );
    }

    protected void getCoords(int index, Complex coords){
        coords.assign(getX(index), getY(index));
    }

    /** assign all complex coords at once */
    protected void assignCoords( Complex [] coords ) {


        boolean propertyHasChanged = coords.length != numOfPoints;
        setNumOfPoints(coords.length, false);

        for( int i=0; i<numOfPoints; i++ ) {
            
            final Complex z = coords[i];
            
            propertyHasChanged = propertyHasChanged || !ComplexProjective1.equals
                ( aRe[i], aIm[i], bRe[i], bIm[i], z.re, z.im);
 
            aRe[i] = z.re;
            aIm[i] = z.im;
            bRe[i] = 1;
            bIm[i] = 0;
        }

	updatePointColor();

        if( propertyHasChanged )
            firePropertyChange( "coords" );
    }

    /* Methods for euclidean coordinates */

    /** set euclidean Coords (x, y) at Index index */
    protected void setCoords (int index, double x, double y ) {
        setProjectiveCoords(index, x, y, 1, 0);
    } 

    protected double getX(int index){
        double a1 = aRe[index];
        double a2 = aIm[index];
        double b1 = bRe[index];
        double b2 = bIm[index];
        double bSqr = b1*b1 + b2*b2;
        
        return ((a1*b1 + a2*b2)/bSqr);
    }

    protected void setX( int anIndex, double x ) {
        double y = getY(anIndex);
	setCoords(anIndex, x, y);
    }

    protected double [] getX() {
	final double[] x = new double[ numOfPoints ];

	for( int i=0; i<numOfPoints; i++) {
	    x[i] = getX(i);
	}

	return x;
    }

    protected void setX( double [] x ) {
	setNumOfPoints( x.length );

	for( int i=0; i<numOfPoints; i++) {
	    setX(i, x[i]);
	}	
	
	firePropertyChange( "x" );
    }

    protected double getY(int index){
        double a1 = aRe[index];
        double a2 = aIm[index];
        double b1 = bRe[index];
        double b2 = bIm[index];
        double bSqr = b1*b1 + b2*b2;
        
        return ((a2*b1 - a1*b2)/bSqr);
    }

    protected void setY( int index, double y ) {
        double x = getX(index);
	setCoords(index, x, y);
    }
    

    protected double [] getY() {
	final double[] y = new double[ numOfPoints ];

	for( int i=0; i<numOfPoints; i++) {
	    y[i] = getY(i);
	}

	return y;
    }

    protected void setY( double [] y ) {
	setNumOfPoints( y.length );

	for( int i=0; i<numOfPoints; i++) {
	    setY(i, y[i]);
	}	
	
	firePropertyChange( "y" );
    }
    
    protected double[] getXY(){
     	final double[] xy = new double[ 2*numOfPoints ];

	for( int i=0; i<numOfPoints; i++) {
	    xy[2*i]    = getX(i);
            xy[2*i +1] = getY(i);
	}

	return xy;
   
    }


    /** assign all euclidean coords at once */
    protected void assignCoords( double[] coords ) {

	if(coords.length % 2 != 0)
	    throw new IllegalArgumentException ("lengths of coords must be multiple of 2");
       
        setNumOfPoints(coords.length / 2, false);

        setNumOfPoints(numOfPoints, false);
        
        for( int i=0, j=0; i<numOfPoints; i++ ) {

            aRe[i] = coords[j++];
            aIm[i] = coords[j++];
            bRe[i] = 1;
            bIm[i] = 0;
        }

	updatePointColor();

	firePropertyChange( "coords" );
    } 


    /* methods for PointColor */
    protected Color [] getPointColor() {
	return pointColor;
    }
    
    protected void setPointColor (Color[] pointColor) {

	if ( pointColor.length != numOfPoints )
	    throw new IllegalArgumentException
                ("lengths of arrays must coinside with the number of points");
        
	this.pointColor = pointColor;

	firePropertyChange( "pointColor" );
    } 

    protected Color getPointColor (int index) {
	return pointColor == null || pointColor[index] == null 
	  ? defaultPointColor : pointColor[index];
    } 

    protected void setPointColor (int index, Color aPointColor) {

	if( pointColor == null )
	    pointColor = new Color[ numOfPoints ];

	if( aPointColor == pointColor[ index ] || 
	    ( aPointColor != null && aPointColor.equals( pointColor[index] ) ) )
	  return;
	
	pointColor[index] = aPointColor;

	firePropertyChange( "pointColor" );
    } 

    protected Color getDefaultPointColor() {
      return defaultPointColor;
    }

    protected void setDefaultPointColor( Color defaultPointColor ) {
      
	if( defaultPointColor == this.defaultPointColor || 
	    ( defaultPointColor != null && defaultPointColor.equals( this.defaultPointColor ) ) )
	  return;
	
	this.defaultPointColor = defaultPointColor;

	firePropertyChange( "defaultPointColor" );
    }

    protected void setColorOfAllPoints ( Color newPointColor ){
        setDoFirePropertyChange(false);

        for (int i=0; i < numOfPoints; i++)
            setPointColor(i, newPointColor);

        setDoFirePropertyChange(true);
    }


    /* draw from interface shape */
    public void draw (MoebiusGraphics g) {
	super.draw (g);
    } 


    protected void drawPoints(MoebiusGraphics g) {
             
        for (int i = 0;i < numOfPoints;i++) {
            
            g.setPointColor( getPointColor( i ) );               
            g.point( aRe[i], aIm[i], bRe[i], bIm[i] );
            
            if( labelPointsWithIndex ){
                g.setVerticalTextLayout( Attributes.TOP );
                g.setHorizontalTextLayout( Attributes.LEFT );
                g.text( aRe[i], aIm[i], bRe[i], bIm[i], " "+Integer.toString( i ) );   
            }       
        }                                       
    }

}

