/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import java.awt.Color;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.Serializable;
import java.util.Vector;

import de.jtem.moebiusViewer.MoebiusGraphics;
import de.jtem.moebiusViewer.MoebiusShape;

public abstract class AbstractShape
    implements Serializable, Cloneable, MoebiusShape {

    private static final long   serialVersionUID = 1937562236303735123L;
    protected PropertyChangeSupport
        propertyChangeSupport = new PropertyChangeSupport( this );
    protected MoebiusShapeChangeSupport
        moebiusShapeChangeSupport = new MoebiusShapeChangeSupport( this );

    public static final  double EPS    = 1e-14;
    public static final  double EPSSQR = EPS*EPS;

    Color                       color = Color.black;
    String                      label = "";
    String                      fontLabel = "";
    boolean                     showLabel = false;
    boolean                     filled = false;
    int                         pointOutline = 1;
    int                         pointRadius = 3;
    double                      lineWidth = 1;
    boolean                     pickable = true;
    boolean                     doFirePropertyChange = true;
    private boolean             changed;


    protected Vector            toolList = new Vector();

    public AbstractShape() {
	//toolList.add( new ShowLabelTool() );
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeSupport.addPropertyChangeListener( listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        propertyChangeSupport.removePropertyChangeListener( listener );
    }

    public void firePropertyChange( String propertyName ) {
        firePropertyChange( propertyName, null, null);
   }

    public void firePropertyChange( String propertyName, Object newValue,
				    Object oldValue ) {
        if(doFirePropertyChange){
            propertyChangeSupport.firePropertyChange( propertyName, newValue,
						  oldValue );
            return;
        }

        changed = true;
    }

    public boolean isDoFirePropertyChange(){
        return doFirePropertyChange;
    }

    public void setDoFirePropertyChange(boolean b){
        if (doFirePropertyChange == b)
            return;
        doFirePropertyChange = b;
        if (b && changed){
            firePropertyChange("changed");
            return;
        }
        changed = false;
    }

    public void dontFirePropertyChange(){
        setDoFirePropertyChange(false);
    }

    public void addMoebiusShapeChangeListener(MoebiusShapeChangeListener listener) {
        moebiusShapeChangeSupport.addMoebiusShapeChangeListener( listener);
    }

    public void removeMoebiusShapeChangeListener(MoebiusShapeChangeListener listener) {
          moebiusShapeChangeSupport.removeMoebiusShapeChangeListener( listener);
      }

    public void fireMoebiusShapeChange( int mask ) {
        moebiusShapeChangeSupport.fireMoebiusShapeChanged( mask );
      }

      public void fireMoebiusShapeChange( ) {
             moebiusShapeChangeSupport.fireMoebiusShapeChanged(0xffff);
           }

    public String getString () {
	return fontLabel;
    }

    public boolean isPickable(){
        return pickable;
    }

    public boolean getPickable(){
        return pickable;
    }

    public void setPickable(boolean b) {
        if (pickable == b)
            return;
        pickable = b;
        firePropertyChange( "pickable" );
    }


  /**
   * Get the value of color.
   * @return Value of color.
   */

  public Color getColor () {
    return color;
  }

  /**
   * Set the value of color.
   * @param v  Value to assign to color.
   */

  public void setColor (Color v) {
      if( color.equals(v) )
	  return;
      color = v;
      firePropertyChange( "color" );
  }

    /**
     * Get the value of label.
     * @return Value of label.
     */

    public String getLabel () {
	return label;
    }

  /**
   * Set the value of label.
   * @param v  Value to assign to label.
   */

  public void setLabel (String v) {
      if( this.label.equals(v) )
	  return;
      this.label = v;
      firePropertyChange( "label" );
  }

  /**
   * Get the value of showLabel.
   * @return Value of showLabel.
   */

  public boolean getShowLabel () {
    return showLabel;
  }

  /**
   * Set the value of showLabel.
   * @param v  Value to assign to showLabel.
   */

  public void setShowLabel (boolean v) {
      if( showLabel == v )
	  return;
      this.showLabel = v;
      firePropertyChange( "showLabel" );
  }

  /**
   * Get the value of pointRadius.
   * @return Value of pointRadius.
   */

  public int getPointRadius () {
    return pointRadius;
  }

  /**
   * Set the value of pointRadius.
   * @param v  Value to assign to pointRadius.
   * @param v < 0 means to use the default provided by the viewer
   */

  public void setPointRadius (int v) {
      if( pointRadius == v )
	  return;
      this.pointRadius = v;
      firePropertyChange( "pointRadius" );
  }

    /** Get the value of pointOutline. @return Value of pointOutline. */

    public int getPointOutline () {
	return pointOutline;
    }

    /** Set the value of pointOutline.  @param v  Value to assign to
	pointOutline. */

    public void setPointOutline (int v) {
	if( pointOutline == v )
	    return;
	pointOutline = v;
	firePropertyChange( "pointOutline" );
    }

    /** Get the value of lineWidth. @return Value of lineWidth. */

    public double getLineWidth () {
	return lineWidth;
    }

    /** Set the value of lineWidth. @param v  Value to assign to lineWidth. */

    public void setLineWidth( double v ) {
	if (equals(lineWidth,v))
	    return;
	lineWidth = v;
	firePropertyChange( "lineWidth" );
    }

  /**
   * Get the value of filled.
   * @return Value of filled.
   */

  public boolean getFilled () {
    return filled;
  }

  /**
   * Set the value of filled.
   * @param v  Value to assign to filled.
   */

  public void setFilled (boolean v) {
      if( filled == v )
	  return;
      this.filled = v;
      firePropertyChange( "filled" );
  }

  public Vector getTools () {

      return pickable ? toolList : null;
  }

  public void reset (MoebiusGraphics context) {
    context.setColor (color);
    context.setPointOutline(pointOutline);
    context.setPointRadius (pointRadius);
    context.setLineWidth(lineWidth);
    context.setFilled (filled);
  }

 public void draw (MoebiusGraphics context) {
     reset(context);
 }

 public Object clone() throws CloneNotSupportedException {
     return super.clone();
 }

 public static boolean equals(double x1, double x2){
     return (Math.abs(x1 - x2) < EPS);
 }
}
