/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import de.jtem.mfc.field.Complex;
import de.jtem.moebiusViewer.MoebiusGraphics;
import de.jtem.moebiusViewer.tool.GridDragVertexTool;

/** A Grid is a 2-dimensional quadrilateral mesh. In order to handle families
    of closed curves, there is a flag <PRE>isClosed</PRE> to determine
    whether the rows are to be considered as closed curves (we saw no need for
    doubly periodic grids) */

public class Grid extends AbstractShape {
    private static final long     serialVersionUID = 1L;

    int numRows = 1;
    int numColums = 1;

    public double[][] re = new double[1][1];
    public double[][] im = new double[1][1];

    boolean isClosed = false;
    boolean drawHorizontalLines = true;
    boolean drawVerticalLines = true;
    boolean drawPoints = true;

    /** Creates a Grid with 1 row, 1 colum, located at the zero */

    public Grid() {
        super();
	toolList.addElement( new GridDragVertexTool() );	
    }

    /** Creates a Grid with integer coordinates, lower left corner at (0,0),
	upper right corner at (m,n) */

    public Grid( int m, int n ) {
        this();
	setNumRows(m);
	setNumColums(n);
    }

    /** Returns the number of Rows */

    public int getNumRows() {
	return numRows;
    }

    /** Change the number of rows, by truncation if the number of rows is
	getting smaller, by linear extrapolation if it is getting bigger.
	<PRE>setNumRows(n)</PRE> is equivalent to <PRE>setNumRows(1)</PRE> if
	n is less than 1. */

    public void setNumRows ( int newNumRows)     {
	newNumRows = Math.max(1 , newNumRows);
	int i, j;
	double baseRe, baseIm, vecRe, vecIm;
	if( newNumRows != numRows ) {

	    double[][] newRe = new double[newNumRows][numColums];
	    double[][] newIm = new double[newNumRows][numColums];

	    int minNumRows = newNumRows < numRows ? newNumRows : numRows;
	    System.arraycopy( re, 0, newRe, 0, minNumRows );
	    System.arraycopy( im, 0, newIm, 0, minNumRows );
	    for( j=0; j<numColums; j++ ) {
		if (newNumRows > numRows) {
		    baseRe = re[numRows -1][j];
		    baseIm = im[numRows -1][j];
		    if (numRows < 2) {
			vecRe = 0;
			vecIm = 1;
		    } else {
			vecRe = re[numRows -1][j] - re[numRows -2][j];
			vecIm = im[numRows -1][j] - im[numRows -2][j];
		    }
		    for (i=0; i<newNumRows - numRows; i++) {
			newRe[numRows + i][j] = baseRe + (i+1) * vecRe;
			newIm[numRows + i][j] = baseIm + (i+1) * vecIm;
		    }
		}
	    }
	    numRows = newNumRows;
	    re = newRe;
	    im = newIm;
	}
    }

    /** Returns the number of colums */

    public int getNumColums() {
	return numColums;
    }

    /** Change the number of colums, by truncation if the number of colums is
	getting smaller, by linear extrapolation if it is getting bigger.
	<PRE>setNumColums(n)</PRE> is equivalent to
	<PRE>setNumRows(1)</PRE> if n is less than 1. */

    public void setNumColums ( int newNumCols) {
	newNumCols = Math.max(1 , newNumCols);
	int i, j;
	double baseRe, baseIm, vecRe, vecIm;
	if( newNumCols != numColums ) {

	    double[][] newRe = new double[numRows][newNumCols];
	    double[][] newIm = new double[numRows][newNumCols];

	    int minNumCols = newNumCols < numColums ? newNumCols : numColums;

	    for( i=0; i<numRows; i++ ) {
		System.arraycopy( re[i], 0, newRe[i], 0, minNumCols );
		System.arraycopy( im[i], 0, newIm[i], 0, minNumCols );
		if (newNumCols > numColums) {
		    if (numColums == 0) {
			baseRe = -1;
			baseIm = 0;
		    } else {
			baseRe = re[i][numColums - 1];
			baseIm = im[i][numColums - 1];
		    }
		    if (numColums < 2) {
			vecRe = 1;
			vecIm = 0;
		    } else {
			vecRe = re[i][numColums -1] - re[i][numColums -2];
			vecIm = im[i][numColums -1] - im[i][numColums -2];
		    }
		    for (j=0; j<newNumCols - numColums; j++) {
			newRe[i][numColums + j] = baseRe + (j+1) * vecRe;
			newIm[i][numColums + j] = baseIm + (j+1) * vecIm;
		    }
		}
	    }
	    re = newRe;
	    im = newIm;
	    numColums = newNumCols;
	}
    }

    public Complex[] getRow( int i) {
	Complex[] row = new Complex[numColums];
	for (int j=0; j<numColums; j++) {
	    row[j] = new Complex(re[i][j], im[i][j]);
	}
	return row;
    }

    public Complex[] getInterpolatedRow(double t) {
	int i = (int) Math.floor(t);
	double s = t - i;
	if (i < 0) {
	    return getRow(0);
	}
	if (i >= numRows -1 ) {
	    return getRow(numRows - 1);
	}
	Complex[] row = new Complex[numColums];
	for (int j=0; j<numColums; j++) {
	    row[j] = new Complex((1-s)*re[i][j] + s*re[i+1][j],
				 (1-s)*im[i][j] + s*im[i+1][j]);
	}
	return row;
    }

    /** First sets the number of colums to the length of <PRE>pts</PRE>.
	If <PRE>i</PRE> is the index of an existing row, pts is copied to the
	coordinates of the i'th row. Otherwise, the number of rows is first
	set to <PRE>i+1</PRE>. If <PRE>i</PRE> is larger than
	<PRE>numRows+1</PRE>, the in-between rows are filled by linear
	interpolation. <PRE>setRow(i,pts)</PRE> does nothing for
	negative i or arrays pts of length 0. */

    public void setRow(int i, Complex[] pts) {
	if (pts.length > 0 && i>=0) {
	    int newNumRows = numRows;
	    int oldNumRows = numRows;
	    setNumColums(pts.length);
	    if (i > numRows - 1) {
		newNumRows = i+1;
		double[][] newRe = new double[newNumRows][numColums];
		double[][] newIm = new double[newNumRows][numColums];
		System.arraycopy( re, 0, newRe, 0, numRows );
		System.arraycopy( im, 0, newIm, 0, numRows );
		re = newRe;
		im = newIm;
	    }
	    for (int j=0; j<numColums; j++) {
		re[i][j] = pts[j].re;
		im[i][j] = pts[j].im;
	    }
	    numRows = newNumRows;
	    makeRuled(oldNumRows - 1, numRows - 1);
	}
    }

    /** Does nothing unless <PRE>pts</PRE> are existing rows. The in-between
	rows are set by linear interpolation, so the vertical curves are ruled
	(straight) in this region. */

    public void makeRuled(int startRow, int endRow) {
	if (startRow >= 0 &&
	    Math.abs(endRow -startRow) > 1 &&
	    endRow < numRows) {
	    int i,j;
	    if (startRow > endRow) {
		i = startRow;
		startRow = endRow;
		endRow = i;
	    }
	    double baseRe,  baseIm, vecRe, vecIm;
	    double factor = 1/((double)(endRow -startRow));
	    for (j=0; j<numColums; j++) {
		baseRe = re[startRow][j];
		baseIm = im[startRow][j];
		vecRe = (re[endRow][j] - baseRe) * factor;
		vecIm = (im[endRow][j] - baseIm) * factor;
		for (i=startRow +1 ; i<endRow; i++) {
		    re[i][j] = baseRe + (i - startRow) * vecRe;
		    im[i][j] = baseIm + (i - startRow) * vecIm;
		}
	    }
	}
    }

    /** Equivalent to <PRE>makeRuled(0, numRows - 1)</PRE>. Makes the whole
	grid ruled*/

    public void makeRuled() {
	makeRuled(0, numRows - 1);
    }

    /** Returns the boolean that tells whether horizontal curves are to be
	treated as closed */

    public boolean getIsClosed() {
	return isClosed;
    }

    /** Sets the boolean that tells whether horizontal curves are to be
	treated as closed */

    public void setIsClosed(boolean  v) {
	this.isClosed = v;
    }

    /** Returns the coordinates in the intersection of the integer rectangle
	<PRE>[i0,i1] x [j0,j1]<PRE> with the grids own index rectangle
	as a freshly created Field.Complex array */

    public Complex[][] getProjectiveCoords(int i0, int j0, int i1, int j1) {
	int startRow = Math.max(0,i0);
	int endRow = Math.min(i1, numRows-1);
	int startColum = Math.max(0,j0);
	int endColum = Math.min(j1,numColums -1);
	int a = Math.max(0,endRow - startRow + 1 );
	int b = Math.max(0,endColum - startColum + 1 );
	int i,j;
	Complex[][] pts = new Complex[a][b];
	for (i=0; i<a; i++) {
	    for (j=0; j<b; j++) {
		pts[i][j] = new Complex(re[i0+i][j0+j], im[i0+i][j0+j]);
	    }
	}
	return pts;
    }

    /** Returns the coordinates as a freshly created Field.Complex array */

    public Complex[][] getCoords() {
	Complex[][] pts = new Complex[numRows][numColums];
	int i,j;
	for (i=0; i<numRows; i++) {
	    for (j=0; j<numColums; j++) {
		pts[i][j] = new Complex(re[i][j], im[i][j]);
	    }
	}
	return pts;
    }

    /** writes the coordinates between row <PRE>i0<PRE> and
	row <PRE>i0+pts.length</PRE> and between colum <PRE>j0</PRE> and
       <PRE>j0+pts[0].length</PRE> into the array <PRE>pts</PRE>. Produces
       index out of bound exceptions if some of these indices are invalid */

    public void getCoords(Complex[][] pts, int i0, int j0) {
	int i,j;
	for (i=0; i<pts.length; i++) {
	    for (j=0; j<pts[0].length; j++) {
		if (pts[i][j] == null) {
		    pts[i][j] = new Complex(re[i0+i][j0+j], im[i0+i][j0+j]);
		} else {
		    pts[i][j].assign(re[i0+i][j0+j], im[i0+i][j0+j]);
		}
	    }
	}
    }

    /** Returns the coordinate in row <PRE>i</PRE>, colum <PRE>j</PRE> as a
	freshly created Field.Complex */

    public Complex getCoords( int i, int j ) {
	Complex z = new Complex();
	getProjectiveCoords( i, j, z );
	return z;
    }

    /** Writes the coordinate in row <PRE>i</PRE>, colum <PRE>j</PRE> into the
	Field.Complex number z. Will produce index out of bound exceptions if
	either <PRE>i</PRE> does not refer to an existing row or if the grid
	is not closed and <PRE>j</PRE> points to a non-existing colum */

    public void getProjectiveCoords( int i, int j, Complex z ) {
	if( isClosed )
	    z.assign( re[i][j % numColums], im[i][j % numColums]);
	else
	    z.assign( re[i][j            ], im[i][j            ]);
    }

    /** Sets the coordinate in row <PRE>i</PRE>, colum <PRE>j</PRE> Will
	produce index out of bound exceptions if
	either <PRE>i</PRE> does not refer to an existing row or if the grid
	is not closed and <PRE>j</PRE> points to a non-existing colum */

    public void setProjectiveCoords( int i, int j, double a, double b ) {
	if( isClosed ) {
	    re[i][j%numColums] = a;
	    im[i][j%numColums] = b;
	} else {
	    re[i][j          ] = a;
	    im[i][j          ] = b;
	}
	firePropertyChange( "coords" );
    }

    /** Sets the coordinate in row <PRE>i</PRE>, colum <PRE>j</PRE> Will
	produce index out of bound exceptions if
	either <PRE>i</PRE> does not refer to an existing row or if the grid
	is not closed and <PRE>j</PRE> points to a non-existing colum */

    public void setProjectiveCoords(int i, int j, Complex p) {
	setProjectiveCoords(i, j, p.re, p.im);
    }

    /** Sets the dimensions of the grid to the dimensions of pts and uses
	the entries of pts (as far as they are non-null) as coordinates. */

    public void setCoords(Complex[][] pts) {
	setNumRows(pts.length);
	setNumColums(pts[0].length);
	int i,j;
	Complex z;
	for (i=0; i<numRows; i++) {
	    for (j=0; j<numColums; j++) {
		z = pts[i][j];
		if (z != null) {
		    re[i][j] = z.re;
		    im[i][j] = z.im;
		}
	    }
	}
	firePropertyChange( "coords" );
    }

    /** Draws the grid */

    public void draw (MoebiusGraphics context) {
	super.draw( context );
	int i,j;
	if (drawHorizontalLines) {
	    for (i=0; i<numRows; i++) {
		for( j=1; j<numColums; j++ )
		    context.lineSeg(re[i][j-1],im[i][j-1],re[i][j],im[i][j]);
		if( isClosed )
		    context.lineSeg(re[i][numColums-1],
				    im[i][numColums-1],
				    re[i][0],
				    im[i][0]);
	    }
	}
	if (drawVerticalLines) {
	    for (i=1; i<numRows; i++) {
		for (j=0; j<numColums; j++)
		    context.lineSeg(re[i-1][j],im[i-1][j],re[i][j],im[i][j]);
	    }
	}
	if (drawPoints) {
	    for (i=0; i<numRows; i++ ) {
		for (j=0; j<numColums; j++)
		    context.point( re[i][j], im[i][j] );
	    }
	}
    }

    /** Returns the boolean that determines whether the horizontal lines
	are drawn */

    public boolean getDrawHorizontalLines() {
	return drawHorizontalLines;
    }

    /** Sets the boolean that determines whether the horizontal lines
	are drawn */

    public void setDrawHorizontalLines(boolean b) {
	drawHorizontalLines = b;
    }

    /** Returns the boolean that determines whether the vertical lines
	are drawn */

    public boolean getDrawVerticalLines() {
	return drawVerticalLines;
    }

    /** Sets the boolean that determines whether the vertical lines
	are drawn */

    public void setDrawVerticalLines(boolean b) {
	drawVerticalLines = b;
    }

    /** Returns the boolean that determines whether the points are drawn */

    public boolean getDrawPoints() {
	return drawPoints;
    }

    /** Sets the boolean that determines whether the points are drawn */

    public void setDrawPoints(boolean b) {
	drawPoints = b;
    }
}
