/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import de.jtem.moebiusViewer.MoebiusGraphics;

public class IndexedLineSet extends PointSet {
  private static final long     serialVersionUID = 1L;

  protected int[] index;
  protected int[] length;


  public IndexedLineSet (){
    super();
  }

  public IndexedLineSet(int n){
    super(n);
  }

  public IndexedLineSet (double[] aXY) {
    super (aXY);
  }

  public IndexedLineSet (double[] aXY, int[] anIndex, int[] aLength) {
    super (aXY);
    setIndexAndLength (anIndex, aLength);
  }

  public boolean isConsistent(){
    return super.isConsistent();
  }

  public int[] getIndex () {
    return index;
  }

  public int[] getLength () {
    return length;
  }

  public int getNumOfComponents () {
    return length != null ? length.length : 0;
  }

  /** @deprecated */
  public void set (int[] anIndex, int[] aLength) {
    setIndexAndLength(anIndex, aLength);
  }

  public void  setIndexAndLength(int[] anIndex, int[] aLength){

    index = anIndex;
    length = aLength;
    checkConsistency();
    firePropertyChange("index and length");
  }

  public void setNumOfPoints (int newNumOfPoints) {

    super.setNumOfPoints(newNumOfPoints);
    checkConsistency();
  }

  /** checks consistency of the current data and stores the result in consistent */
  protected void checkConsistency(){

    boolean c = true;
    if (!(index == null || length == null)){
      for (int i = 0; i<index.length; i++){
	if (c)
	  c = index[i] < numOfPoints ? true : false;
	if (c)
	  c = index[i] >= 0 ? true : false;
      }
      if (c){
	int l = 0;
	for (int i = 0; i<length.length; i++){
	  l += length[i];
	}
	c = l <= index.length ? true : false;
      }
    }
    setConsistent(c);
  }

  public void draw (MoebiusGraphics g) {
    if (consistent){
      super.draw (g);
    }
  }

  protected void drawPoints(MoebiusGraphics g){
    int numOfComponents = length.length;

    for (int i = 0, component = 0;component < numOfComponents; component++) {
      int       endOfComponent = i + length[component];
      double    aR1 = aRe[index[i]];
      double    aI1 = aIm[index[i]];
      double    bR1 = bRe[index[i]];
      double    bI1 = bIm[index[i]];

      for (;i < endOfComponent;i++) {
	double    aR2 = aRe[index[i]];
	double    aI2 = aIm[index[i]];
	double    bR2 = bRe[index[i]];
	double    bI2 = bIm[index[i]];
	g.lineSeg (aR1, aI1, bR1, bI1, aR2, aI2, bR2, bI2);
	aR1 = aR2;
	aI1 = aI2;
	bR1 = bR2;
	bI1 = bI2;
      }
    }
    super.drawPoints(g);
  }

  void reverseComponent( int comp, int startIndex ) {

    int endIndex   = startIndex + length[ comp ] - 1;
    int midIndex   = ( startIndex + endIndex + 1) / 2;

    for( int i=startIndex, j=endIndex; i<midIndex; i++, j-- ) {
      int tmp = index[i];
      index[i] = index[j];
      index[j] = tmp;
    }
  }

  int getStartIndexOfComponent( int comp ) {
    int start = 0;
    for( int i=0; i<comp; i++ )
      start += length[i];
    return start;
  }

  public void reverseComponent( int comp ) {
    reverseComponent( comp, getStartIndexOfComponent( comp ) );
  }

  public Polygon getComponent( int comp ) {

    int startIndex = getStartIndexOfComponent( comp );
    int endIndex   = startIndex + length[ comp ] - 1;

    Polygon polygon = new Polygon();

    polygon.closed = index[ startIndex ] == index[ endIndex ];

    int nop = length[ comp ] - ( polygon.closed ? 1 : 0 );

    polygon.setNumOfPoints( nop, false );

    for( int i=0, j=startIndex; i<nop; i++, j++ ) {

      polygon.aRe[i] = aRe[ index[ j ] ];
      polygon.aIm[i] = aIm[ index[ j ] ];
      polygon.bRe[i] = bRe[ index[ j ] ];
      polygon.bIm[i] = bIm[ index[ j ] ];
    }

    return polygon;
  }

  class Close {

    int [] startIndex = new int[ length.length ];

    int [] start = new int[ length.length ];
    int [] end   = new int[ length.length ];

    int [] next  = new int[ length.length ];
    int [] prev  = new int[ length.length ];
    int [] comp  = new int[ length.length ];

    Close() {

      for( int i=0, J=0; i<length.length; i++ ) {
	next[i] = -1;
	prev[i] = -1;
	comp[i] = -1;

	startIndex[i] = J;

	start[i] = index[J];    J += length[i];
	end[i]   = index[J-1];
      }

    }

    private int sortComponent( int i, int compIndex ) {

      comp[i] = compIndex;

      // //System.out.println(  "sort comp " + compIndex + "org comp = " + i );
      // sort forward
      for( int runner = i; next[ runner ] != -1 && comp[ next[ runner ] ] == -1; runner = next[ runner ] ) {

	final int nextOfRunner = next[ runner ];

	comp[ nextOfRunner ] = compIndex;

	//System.out.println(  "sort comp " + compIndex + " -> org comp = " + nextOfRunner );

	if( end[ runner ] == start[ nextOfRunner ] )
	  continue;

	reverseComponent( nextOfRunner, startIndex[ nextOfRunner ] );

	int tmp = prev[ nextOfRunner ];
	prev[ nextOfRunner ] = next[ nextOfRunner ];
	next[ nextOfRunner ] = tmp;

	tmp = start[ nextOfRunner ];
	start[ nextOfRunner ] = end[ nextOfRunner ];
	end  [ nextOfRunner ] = tmp;

      }

      // sort backward
      int runner = i;

      while( prev[ runner ] != -1  && comp[ prev[ runner ] ] == -1 ) {

	final int prevOfRunner = prev[ runner ];

	comp[ prevOfRunner  ] = compIndex;

	//System.out.println(  "sort comp " + compIndex + " <- org comp = " + prevOfRunner );

	if( start[ runner ] == end[ prevOfRunner ] ) {
	  runner = prevOfRunner;
	  continue;
	}

	reverseComponent( prevOfRunner, startIndex[ prevOfRunner ] );

	int tmp = prev[ prevOfRunner ];
	prev[ prevOfRunner ] = next[ prevOfRunner ];
	next[ prevOfRunner ] = tmp;


	tmp = start[ prevOfRunner ];
	start[ prevOfRunner ] = end[ prevOfRunner ];
	end  [ prevOfRunner ] = tmp;

	runner = prevOfRunner;
      }

      return runner;
    }


    void perform() {

      for( int i=0; i<length.length; i++ ) {

	final int s = start[i];
	final int e = end  [i];

	if( next[i] == -1 )
	  for( int j=i+1; j<length.length; j++ ) {
	    if( prev[j] == -1 && start[j] == e ) {
	      next[i] = j;
	      prev[j] = i;
	      //System.out.println(  i + "<-> " + j );
	      break;
	    }
	    if( next[j] == -1 && end  [j] == e ) {
	      next[i] = j;
	      next[j] = i;
	      //System.out.println(  i + "<-> " + j );
	      break;
	    }
	  }

	if( prev[i] == -1 )
	  for( int j=i+1; j<length.length; j++ ) {
	    if( prev[j] == -1 && start[j] == s ) {
	      prev[i] = j;
	      prev[j] = i;
	      //System.out.println(  i + "<-> " + j );
	      break;
	    }
	    if( next[j] == -1 && end  [j] == s ) {
	      prev[i] = j;
	      next[j] = i;
	      //System.out.println(  i + "<-> " + j );
	      break;
	    }
	  }

	//System.out.println( i + " next= " + next[i]+ " prev= " + prev[i] );
      }

      int numOfComp   = 0;
      int totalLength = 0;

      int [] tmpNewLength = new int[ length.length ];
      int [] firstOfComp  = new int[ length.length ];

      for( int i=0; i<length.length; i++ ) {

	if( comp[i] != -1 )
	  continue;

	firstOfComp[ numOfComp ] = sortComponent( i, numOfComp );

	//System.out.println( "first " +  firstOfComp[ numOfComp ] );

	tmpNewLength[ numOfComp ] = length[ firstOfComp[ numOfComp ] ];

	int runner = next[ firstOfComp[ numOfComp ] ];
	while( runner != -1 && runner != firstOfComp[ numOfComp ] ) {
	  tmpNewLength[ numOfComp ] += length[ runner ] - 1;
	  runner = next[ runner ];
	}

	totalLength += tmpNewLength[ numOfComp ];

	numOfComp++;
      }

      int [] newLength = new int[ numOfComp ];
      int [] newIndex  = new int[ totalLength ];

      System.arraycopy( tmpNewLength, 0, newLength, 0, numOfComp );

      int J=0;
      for( int i=0; i<numOfComp; i++ ) {
	int runner = firstOfComp[i];

	System.arraycopy( index, startIndex[ runner ], newIndex, J, length[ runner ] );
	//System.out.println( runner + "  " + start[runner] + " -> " + end[runner] + " next= " + next[runner]+ " prev= " + prev[runner] );
	J += length[runner];

	runner = next[runner];

	while( runner != -1 && runner != firstOfComp[i] ) {
	  //System.out.println( runner + "  " + start[runner] + " -> " + end[runner]+ " next= " + next[runner]+ " prev= " + prev[runner] );
	  System.arraycopy( index, startIndex[ runner ] + 1, newIndex, J, length[ runner ] - 1 );
	  /*if( next[runner] != -1 )
	    if( end[runner] != start[ next[runner] ] )
	    throw new RuntimeException( "ups: " ); */
	  J += length[runner] - 1;
	  runner = next[ runner ];
	}

      }

      if( J != newIndex.length )
	throw new RuntimeException( "ups!!! " + J + " <> " + newIndex.length );
      //System.out.println( "ups!!! " + J + " <> " + newIndex.length );

      //new blas.IntegerVector( newIndex ).print( "new index" );
      //new blas.IntegerVector( newLength ).print( "new length" );
      index  = newIndex;
      length = newLength;
    }

  }

  public void close() {
    new Close().perform();
  }

}


