/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import de.jtem.moebiusViewer.MoebiusShape;

/**
 * Support class for mutable shapes. Manages the listener
 * registration and event sending.
 * On serialization all registered listeners which implement
 * {@link java.io.Serializable Serializable} are serialized
 * too.
 * In the MoebiusShape implementor class the support class can
 * be used as follows:<br><pre>
 * ShapeSupport support=<B><FONT color=#804040>new</FONT></B> ShapeSupport(<B><FONT color=#2e8b57>this</FONT></B>);
 * <B><FONT color=#2e8b57>public</FONT></B> <B><FONT color=#2e8b57>void</FONT></B> addShapeListener(ShapeListener l)
 * {
 *   support.addShapeListener(l);
 * }
 * <B><FONT color=#2e8b57>public</FONT></B> <B><FONT color=#2e8b57>void</FONT></B> removeShapeListener(ShapeListener l)
 * {
 *   support.removeShapeListener(l);
 * }
 * <B><FONT color=#2e8b57>protected</FONT></B> <B><FONT color=#2e8b57>void</FONT></B> fireShapeChange(<B><FONT color=#2e8b57>int</FONT></B> mask)
 * {
 *   support.shapeChanged(mask);
 * }
 * </PRE>
 * @see ShapeListener
 * @see ShapeEvent
 * @see MoebisShape
 */
public class MoebiusShapeChangeSupport implements java.io.Serializable
{
//private static final long serialVersionUID=-6143627397035678465l;
  private int numListener;
  private MoebiusShapeChangeListener[] listener;
  private MoebiusShape source;
  private Object key;
  private MoebiusShapeChangeEvent ev;
  public MoebiusShapeChangeSupport(MoebiusShape src)
  {
    source=src;
    numListener=0;
    listener=new MoebiusShapeChangeListener[10];
    key=new Object();
    ev=new MoebiusShapeChangeEvent(source);
  }
  /**
   * Must be called every time the shape has
   * changed. Will send events only if at least one
   * bit of the mask is set.
   */
  public void fireMoebiusShapeChanged(int mask)
  {
    if(mask==0) return;
    synchronized(key)
    {
      ev.setChanges(mask);
      for(int i=0; i<numListener; i++)
        listener[i].moebiusShapeChanged(ev);
    }
  }
  public void addMoebiusShapeChangeListener(MoebiusShapeChangeListener l)
  {
    synchronized(key)
    {
      if(numListener==listener.length)
      {
        MoebiusShapeChangeListener[] newListener=new MoebiusShapeChangeListener[numListener+10];
        System.arraycopy(listener,0,newListener,0,numListener);
        listener=newListener;
      }
      listener[numListener++]=l;
    }
  }
  public void removeMoebiusShapeChangeListener(MoebiusShapeChangeListener l)
  {
    synchronized(key)
    {
      int i;
      find:
      {
        for(i=0; i<numListener; i++)
          if(listener[i]==l) break find;
        return;
      }
      numListener--;
      if(i!=numListener)
      {
        System.arraycopy(listener,i+1,listener,i,numListener-i);
      }
      listener[numListener]=null;
    }
  }
  private void writeObject(java.io.ObjectOutputStream stream)
    throws java.io.IOException
  {
    stream.writeObject(source);
    for(int i=0; i<numListener; i++)
      if(listener[i] instanceof java.io.Serializable)
        stream.writeObject(listener[i]);
    stream.writeObject(null);
  }
  private void readObject(java.io.ObjectInputStream stream)
    throws java.io.IOException, ClassNotFoundException
  {
    source=(MoebiusShape)stream.readObject();
    ev=new MoebiusShapeChangeEvent(source);
    java.util.ArrayList v=new java.util.ArrayList();
    Object o;
    while((o=stream.readObject())!=null)
      v.add(o);
    numListener=v.size();
    listener=new MoebiusShapeChangeListener[numListener<10?10:numListener];
    v.toArray(listener);
    key=new Object();
  }
}
