/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import java.awt.Color;

import de.jtem.mfc.field.Complex;
import de.jtem.mfc.field.Field;
import de.jtem.mfc.geometry.ComplexProjective1;
import de.jtem.moebiusViewer.Attributes;
import de.jtem.moebiusViewer.MoebiusGraphics;
import de.jtem.moebiusViewer.tool.DragPointTool;
import de.jtem.moebiusViewer.tool.ShowLabelTool;

public class Point extends AbstractShape implements Field.Complex {
    private static final long     serialVersionUID = 1L;
    double aRe = 0;
    double aIm = 0;
    double bRe = 1;
    double bIm = 0;

    Color pointColor = Color.red;

    public Point () {
      super();
      toolList.addElement( new DragPointTool() );
      toolList.addElement( new ShowLabelTool() );
    }

    public Point( double x, double y ) {
	this();
	setCoords(x, y);
    }

    public Point( Field.Complex z1 ) {
        this();
        setCoords(z1);
    }
    public Point( Complex z1 ) {
    	this();
    	setCoords(z1);
    }
    public Point( Complex z1, Complex z2) {
        this();
        setProjectiveCoords(z1, z2);
    }

    public Point( ComplexProjective1 aZ){
        this();
        setProjectiveCoords(aZ);
    }

    public double getRe() {
        return getX();
    }

    public double getIm() {
        return getY();
    }

    public void assign(Field.Complex z) {
        setCoords(z.getRe(),z.getIm());
    }

    public Color getPointColor () {
	return pointColor;
    }

    public void setPointColor (Color v) {
      if( pointColor.equals(v) )
	  return;
      pointColor = v;
      firePropertyChange( "pointColor" );

    }

    /* methods for projective Coordinates */

    public ComplexProjective1 getProjectiveCoords(){
        final ComplexProjective1 coords = new ComplexProjective1(aRe, aIm, bRe, bIm);
        return coords;
    }

    public Complex getCoords(){
      Complex c = new Complex( aRe, aIm );
      c.assignDivide( bRe, bIm );
      return c;
    }

    public void getCoords(ComplexProjective1 coords){
        coords.assign(aRe, aIm, bRe, bIm);
    }

    public void setProjectiveCoords (ComplexProjective1 coords){
        setProjectiveCoords(coords.aRe, coords.aIm, coords.bRe, coords.bIm);
    }

    /* This method does the real work and fires a Property Change */
    public void setProjectiveCoords(double aRe1, double aIm1, double bRe1, double bIm1){
        if (ComplexProjective1.equals(aRe, aIm, bRe, bIm, aRe1, aIm1, bRe1, bIm1))
            return;
        aRe = aRe1;
        aIm = aIm1;
        bRe = bRe1;
        bIm = bIm1;
        firePropertyChange("coords");
        fireMoebiusShapeChange(MoebiusShapeChangeEvent.COORDS_CHANGED);
    }

    public void setProjectiveCoords (Complex a, Complex b){
        setProjectiveCoords (a.re, a.im, b.re, b.im);
    }


    /* methods for euclidean Coordinates */

    public void setCoords (double x, double y){
        setProjectiveCoords(x, y, 1, 0);
    }

    public void setCoords (Complex z){
        setCoords(z.re, z.im);
    }
    
    public void setCoords (Field.Complex z){
    	setCoords(z.getRe(), z.getIm() );
    }
    
    public void getCoords (Complex z){
        z.assign(getX(), getY());
    }

    /**
     * Get the value of x.
     * @return Value of x.
     */
    public double getX () {
        double bSqr = bRe*bRe + bIm*bIm;
        return ((aRe*bRe + aIm*bIm)/bSqr);
    }

    /**
     * Set the value of x.
     * @param v  Value to assign to x.
     */
    public void setX( double x) {
        double oldY = getY();
        setCoords(x, oldY);
    }

    /**
     * Get the value of y.
     * @return Value of y.
     */
    public double getY () {
        double bSqr = bRe*bRe + bIm*bIm;
        return ((aIm*bRe - aRe*bIm)/bSqr);
    }


    /**
     * Set the value of y.
     * @param v  Value to assign to y.
     */
    public void setY( double y) {
        double oldX = getX();
        setCoords(oldX, y);
    }

    /** @deprecated */
    public void set( double x, double y ) {
        setCoords(x, y);
    }

    /** @deprecated */
    public void set( Complex z ) {
        setCoords(z);
    }


//     public void reset (MoebiusGraphics context) {
// 	super.reset( context );
// 	context.setPointColor(pointColor);
//     }

    public void draw (MoebiusGraphics g) {
	super.draw (g);
        g.setPointColor(pointColor);
	g.point (aRe, aIm, bRe, bIm);

	if( showLabel ) {
            g.setVerticalTextLayout( Attributes.TOP );
            g.setHorizontalTextLayout( Attributes.LEFT );
            g.text( aRe, aIm, bRe, bIm, " " +label );
	}

    }

    //public String toString () {
    //return "(" + getX() + "," + getY() + ")";
    //}


}










