/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import java.awt.Color;

import de.jtem.mfc.field.Complex;
import de.jtem.mfc.geometry.ComplexProjective1;
import de.jtem.moebiusViewer.MoebiusGraphics;
import de.jtem.moebiusViewer.tool.QuadMeshDragTool;

public class QuadMesh
    extends AbstractPointSet {
    private static final long serialVersionUID = 1L;

    int numOfRows;
    int numOfColumns;
    boolean drawPoints = true;
    boolean drawVerticalLines = true;
    boolean drawHorizontalLines = true;
    boolean closedInXDirection = false;
    boolean closedInYDirection = false;

    public QuadMesh() {
        toolList.add(0, new QuadMeshDragTool());
        assignSquareGrid(2, 2);
    }

    public QuadMesh(int m, int n) {
        this();
        assignSquareGrid(m, n);
    }

    public QuadMesh(double[] coords) {
        this();
        assignCoords(coords);
    }

    public QuadMesh(Complex[] coords) {
        this(coords.length, 1);
        assignCoords(coords);
    }

    public QuadMesh(ComplexProjective1[] coords) {
        this(coords.length, 1);
        assignCoords(coords);
    }

    public boolean isClosedInXDirection() {
        return closedInXDirection;
    }

    public boolean isClosedInYDirection() {
        return closedInYDirection;
    }

    public void setClosedInXDirection(final boolean xClosed) {
        if (closedInXDirection == xClosed) {
            return;
        }
        closedInXDirection = xClosed;
        firePropertyChange("drawPoints");
    }

    public void setClosedInYDirection(final boolean yClosed) {
        if (closedInYDirection == yClosed) {
            return;
        }
        closedInYDirection = yClosed;
        firePropertyChange("drawPoints");
    }

    public boolean isLabelPointsWithIndex() {
        return super.isLabelPointsWithIndex();
    }

    public boolean getLabelPointsWithIndex() {
        return super.isLabelPointsWithIndex();
    }

    public void setLabelPointsWithIndex(boolean aBoolean) {
        super.setLabelPointsWithIndex(aBoolean);
    }

    public boolean isConsistent() {
        return super.isConsistent();
    }

    /**
     * Get the value of numOfRows.
     * @return Value of numRows.
     */

    public int getNumOfRows() {
        return numOfRows;
    }

    /** @deprecated */
    public int getNumRows() {
        return numOfRows;
    }

    /**
     * Set the value of numOfRows.
     * @param v  Value to assign to numOfRows.
     */

    public void setNumOfRows(int v) {
        if (v == numOfRows) {
            return;
        }
        setNumOfRowsAndColumns(v, numOfColumns, false);
    }

    /** @deprecated */
    public void setNumRows(int v) {
        setNumOfRows(v);
    }

    /**
     * Get the value of numOfColumns.
     * @return Value of numOfColumns.
     */

    public int getNumOfColumns() {
        return numOfColumns;
    }

    /** @deprecated */
    public int getNumCols() {
        return numOfColumns;
    }

    /**
     * Set the value of numOfColumns.
     * @param v  Value to assign to numOfColumns.
     */
    public void setNumOfColumns(int v) {
        if (v == numOfColumns) {
            return;
        }
        setNumOfRowsAndColumns(numOfRows, v, false);
        firePropertyChange("numOfColumns");
    }

    /** @deprecated */
    public void setNumCols(int v) {
        setNumOfRowsAndColumns(numOfRows, v);
    }

    /** @deprecated */
    public void setNumRowsAndNumCols(int newNumRows, int newNumCols) {
        setNumOfRowsAndColumns(newNumRows, newNumCols);
    }

    public void setNumOfRowsAndColumns(int m, int n) {
        setNumOfRowsAndColumns(m, n, true);
    }

    protected void setNumOfRowsAndColumns(int m, int n,
                                          boolean doFirePropertyChange) {
        if ( (numOfRows == m) && (numOfColumns == n)) {
            return;
        }

        final int minRows = numOfRows < m ? numOfRows : m;
        final int minColumns = numOfColumns < n ? numOfColumns : n;
        final int newNumOfPoints = m * n;
        double[] newARe = new double[newNumOfPoints];
        double[] newAIm = new double[newNumOfPoints];
        double[] newBRe = new double[newNumOfPoints];
        double[] newBIm = new double[newNumOfPoints];

        for (int i = 0; i < newNumOfPoints; i++) {
            newBRe[i] = 1;
        }

        if (aRe != null) {
            for (int i = 0; i < minRows; i++) {
                int oldRow = i * numOfColumns;
                int newRow = i * n;
                System.arraycopy(aRe, oldRow, newARe, newRow, minColumns);
                System.arraycopy(aIm, oldRow, newAIm, newRow, minColumns);
                System.arraycopy(bRe, oldRow, newBRe, newRow, minColumns);
                System.arraycopy(bIm, oldRow, newBIm, newRow, minColumns);
            }
        }

        aRe = newARe;
        aIm = newAIm;
        bRe = newBRe;
        bIm = newBIm;

        numOfRows = m;
        numOfColumns = n;
        numOfPoints = newNumOfPoints;

        updatePointColor();

        if (doFirePropertyChange) {
            firePropertyChange("numOfRows and numOfColumns");
        }
    }

    public boolean isDrawPoints() {
        return drawPoints;
    }

    public void setDrawPoints(boolean b) {
        if (drawPoints == b) {
            return;
        }
        drawPoints = b;
        firePropertyChange("drawPoints");
    }

    public boolean isDrawVerticalLines() {
        return drawVerticalLines;
    }

    public void setDrawVerticalLines(boolean b) {
        if (drawVerticalLines == b) {
            return;
        }
        drawVerticalLines = b;
        firePropertyChange("drawVerticalLines");
    }

    public boolean isDrawHorizontalLines() {
        return drawHorizontalLines;
    }

    public void setDrawHorizontalLines(boolean b) {
        if (drawHorizontalLines == b) {
            return;
        }
        drawHorizontalLines = b;
        firePropertyChange("drawHorizontalLines");
    }

    /* methods for ComplexProjective coords */

    public void setProjectiveCoords(int i, int j, double aRe1, double aIm1,
                                    double bRe1, double bIm1) {
        super.setProjectiveCoords(i * numOfColumns + j, aRe1, aIm1, bRe1, bIm1);
    }

    public void setProjectiveCoords(int i, int j, final ComplexProjective1 z) {
        setProjectiveCoords(i, j, z.aRe, z.aIm, z.bRe, z.bIm);
    }

    public ComplexProjective1[][] getProjectiveCoords() {
        ComplexProjective1[][] coords = new ComplexProjective1[numOfRows][
            numOfColumns];
        for (int i = 0; i < numOfRows; i++) {
            for (int j = 0; j < numOfColumns; j++) {
                coords[i][j] = getProjectiveCoords(i, j);
            }
        }
        return coords;
    }

    public void setProjectiveCoords(ComplexProjective1[][] coords) {

        try {
            setNumOfRowsAndColumns(coords.length, coords[0].length, false);

            for (int i = 0; i < numOfRows; i++) {
                for (int j = 0; j < numOfColumns; j++) {
                    if (coords[i][j] != null) {
                        setProjectiveCoords(i, j, coords[i][j]);
                    }
                }
            }
        }
        catch (IndexOutOfBoundsException e) {
            throw new IllegalArgumentException("coords array not rectangular");
        }
    }

    public Complex[][] getCoords() {
        Complex[][] coords = new Complex[numOfRows][numOfColumns];
        for (int i = 0; i < numOfRows; i++) {
            for (int j = 0; j < numOfColumns; j++) {
                coords[i][j] = getCoords(i, j);
            }
        }
        return coords;
    }

    public void setCoords(Complex[][] coords) {

        try {
            setNumOfRowsAndColumns(coords.length, coords[0].length, false);

            for (int i = 0; i < numOfRows; i++) {
                for (int j = 0; j < numOfColumns; j++) {
                    if (coords[i][j] != null) {
                        setCoords(i, j, coords[i][j]);
                    }
                }
            }
        }
        catch (IndexOutOfBoundsException e) {
            throw new IllegalArgumentException("coords array not rectangular");
        }
    }

    /** assign all coords at once */
    public void assignCoords(ComplexProjective1[] coords) {
        super.assignCoords(coords);
        checkConsistency();
    }

    public void assignCoords(ComplexProjective1[] coords, int m, int n) {
        setNumOfRowsAndColumns(m, n, false);
        this.assignCoords(coords);
    }

    public void getProjectiveCoords(int i, int j, ComplexProjective1 coords) {
        super.getProjectiveCoords(i * numOfColumns + j, coords);
    }

    public void getCoords(int i, int j, Complex coords) {
        super.getCoords(i * numOfColumns + j, coords);
    }

    public ComplexProjective1 getProjectiveCoords(int i, int j) {
        return super.getProjectiveCoords(i * numOfColumns + j);
    }

    public Complex getCoords(int i, int j) {
        return super.getCoords(i * numOfColumns + j);
    }

    /* methods for Field.Complex Coords */
    public void setCoords(int i, int j, final Complex z) {
        super.setCoords(i * numOfColumns + j, z.re, z.im);
    }

    /** assign all complex coords at once */
    public void assignCoords(Complex[] coords) {
        super.assignCoords(coords);
        checkConsistency();
    }

    public void assignCoords(Complex[] coords, int m, int n) {
        setNumOfRowsAndColumns(m, n);
        this.assignCoords(coords);
    }

    /* Methods for euclidean coordinates */

    /** set euclidean Coords (x, y) at Index (i,j) */
    public void setCoords(int i, int j, double x, double y) {
        setProjectiveCoords(i, j, x, y, 1, 0);
    }

    public double getX(int i, int j) {
        int index = i * numOfColumns + j;
        return super.getX(index);
    }

    public void setX(int i, int j, double x) {
        super.setX(i * numOfColumns + j, x);
    }

    public double[] getX() {
        return super.getX();
    }

    public void setX(double[] x) {
        super.setX(x);
        checkConsistency();
    }

    public double getY(int i, int j) {
        return super.getY(i * numOfColumns + j);
    }

    public void setY(int i, int j, double y) {
        super.setY(i * numOfColumns + j, y);
    }

    public double[] getY() {
        return super.getY();
    }

    public void setY(double[] y) {
        super.setY(y);
        checkConsistency();
    }

    public double[] getXY() {
        return super.getXY();

    }

    /** assign all euclidean coords at once */
    public void assignCoords(double[] coords) {
        super.assignCoords(coords);
        checkConsistency();
    }

    public void assignCoords(double[] coords, int m, int n) {
        setNumOfRowsAndColumns(m, n);
        this.assignCoords(coords);
    }

    /** @deprecated */
    public void set(int i, int j, double x, double y) {
        setCoords(i, j, x, y);
    }

    /* methods for PointColor */
    public Color[] getPointColor() {
        return super.getPointColor();
    }

    public void setPointColor(Color[] pointColor) {
        super.setPointColor(pointColor);
    }

    public Color getPointColor(int i, int j) {
        return super.getPointColor(i * numOfColumns + j);
    }

    public void setPointColor(int i, int j, Color aPointColor) {
        super.setPointColor(i * numOfColumns + j, aPointColor);
    }

    public Color getDefaultPointColor() {
        return super.getDefaultPointColor();
    }

    public void setDefaultPointColor(Color color) {
        super.setDefaultPointColor(color);
    }

    /** checks the consistency of the current data and stores the result in consistent */
    protected void checkConsistency() {
        boolean c = numOfRows * numOfColumns == getNumOfPoints() ? true : false;
        setConsistent(c);
    }

    public void assignSquareGrid(int n, int m) {

        setNumOfRowsAndColumns(n, m, false);
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < m; j++) {
                setProjectiveCoords(i, j, (double) i, (double) j, 1, 0);
            }
        }
        setConsistent(true);
    }

    /* draw from interface shape */
    public void draw(MoebiusGraphics g) {

        if (consistent) {
            super.draw(g);
            if (lineWidth != 0) {
                if (drawHorizontalLines && lineWidth != 0) {
                    for (int i = 0; i < numOfRows; i++) {
                        for (int j = 1; j < numOfColumns; j++) {
                            int row = i * numOfColumns;
                            int col = j;

                            double aR1 = aRe[row + col - 1];
                            double aI1 = aIm[row + col - 1];
                            double bR1 = bRe[row + col - 1];
                            double bI1 = bIm[row + col - 1];
                            double aR2 = aRe[row + col];
                            double aI2 = aIm[row + col];
                            double bR2 = bRe[row + col];
                            double bI2 = bIm[row + col];
                            g.lineSeg(aR1, aI1, bR1, bI1, aR2, aI2, bR2, bI2);
                        }
                    }
                    if (closedInXDirection) {
                        for (int i = 0; i < numOfRows; i++) {
                            int row = i * numOfColumns;

                            double aR1 = aRe[row];
                            double aI1 = aIm[row];
                            double bR1 = bRe[row];
                            double bI1 = bIm[row];
                            double aR2 = aRe[row + numOfColumns - 1];
                            double aI2 = aIm[row + numOfColumns - 1];
                            double bR2 = bRe[row + numOfColumns - 1];
                            double bI2 = bIm[row + numOfColumns - 1];
                            g.lineSeg(aR1, aI1, bR1, bI1, aR2, aI2, bR2, bI2);
                        }
                    }
                }
                if (drawVerticalLines) {
                    for (int i = 1; i < numOfRows; i++) {
                        for (int j = 0; j < numOfColumns; j++) {
                            int row = i * numOfColumns;
                            int col = j;

                            double aR1 = aRe[row - numOfColumns + col];
                            double aI1 = aIm[row - numOfColumns + col];
                            double bR1 = bRe[row - numOfColumns + col];
                            double bI1 = bIm[row - numOfColumns + col];
                            double aR2 = aRe[row + col];
                            double aI2 = aIm[row + col];
                            double bR2 = bRe[row + col];
                            double bI2 = bIm[row + col];
                            g.lineSeg(aR1, aI1, bR1, bI1, aR2, aI2, bR2, bI2);
                        }
                    }
                    if (closedInYDirection) {
                        int row = (numOfRows - 1) * numOfColumns;
                        for (int j = 0; j < numOfColumns; j++) {
                            int col = j;

                            double aR1 = aRe[col];
                            double aI1 = aIm[col];
                            double bR1 = bRe[col];
                            double bI1 = bIm[col];
                            double aR2 = aRe[row + col];
                            double aI2 = aIm[row + col];
                            double bR2 = bRe[row + col];
                            double bI2 = bIm[row + col];
                            g.lineSeg(aR1, aI1, bR1, bI1, aR2, aI2, bR2, bI2);
                        }
                    }
                }
            }
            if (drawPoints) {
                drawPoints(g);
            }

        }
    }

}
