/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.shape;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Vector;

import de.jtem.moebiusViewer.MoebiusGraphics;
import de.jtem.moebiusViewer.MoebiusShape;
import de.jtem.moebiusViewer.RenderTreeInfo;

public class ShapeSet extends AbstractShape implements PropertyChangeListener {
  private static final long     serialVersionUID = 1L;

  protected Vector              shapes = new Vector ();
  
  public void propertyChange( PropertyChangeEvent e ) {
    
    firePropertyChange( e.getPropertyName(), e.getOldValue(), e.getNewValue() );
  }

  /**
   * Get the Vector of Shapes.
   * @return Vector of Shapes.
   */

  public Vector getShapes () {
    return shapes;
  } 

  public ShapeSet () {
  }
    
  private static boolean contains( MoebiusShape[] shapes, MoebiusShape test ) {
    if( test == null )
      return true;

    for( int i=0; i<shapes.length; i++ )
      if( shapes[i] == test )
	return true;
    
    return false;
  }

  public void set( MoebiusShape[] newShapes ) {
    if( newShapes == null ) {
      removeAll();	    
    } 
    else {
      for( int i=0; i<newShapes.length; i++ )
	if( !shapes.contains( newShapes[i] ) )
	  add( newShapes[i] );
      
      for( int i=0; i<shapes.size(); i++ ) {
	MoebiusShape aShape = (MoebiusShape)shapes.elementAt(i);
	
	if( !contains( newShapes, aShape ) )
	  remove( aShape );
      }
    }
  }

  public void set( int index, MoebiusShape newShape ) {
    if( newShape == null ) {
        throw new NullPointerException( "shape is null" );	    
    } 

    if( shapes.elementAt( index ) == newShape )
        return;

    int oldIndex = shapes.indexOf( newShape );

    if( oldIndex != -1 ) {
        shapes.set( oldIndex, null );
        shapes.set(    index, newShape );
        return;
    }

    shapes.set( index, newShape );
   
    try {
      newShape.getClass().getMethod( "addPropertyChangeListener", new Class[] { PropertyChangeListener.class } )
	.invoke( newShape, new Object[] { this } );
    }
    catch( Exception e ) { 
    }

    firePropertyChange( "added shape", newShape, null );
  }

  /**
   * Adds a shape to the set of shapes.
   * @param <{de.jtem.moebiusViewer.MoebiusShape}>
   */

  public void add (MoebiusShape aShape) {
    
    if( RenderTreeInfo.contains( aShape, this ) )
      throw new IllegalArgumentException(" generated loop in render tree" );
    
    shapes.addElement (aShape);

    try {
      aShape.getClass().getMethod( "addPropertyChangeListener", new Class[] { PropertyChangeListener.class } )
	.invoke( aShape, new Object[] { this } );
    }
    catch( Exception e ) { 
    }
	  
    firePropertyChange( "added shape", aShape, null );
  } 

  /**
   * Adds an array of shapes to the set
   * @param Shape2D []
   */

  public void add (MoebiusShape[] shape) {
    for (int i = 0;i < shape.length;i++) {
      add (shape[i]);
    } 
  } 

  /**
   * Removes the shape @param Shape2D from
   * the shape set.
   * @param <{de.jtem.moebiusViewer.MoebiusShape}>
   * @return true if the argument was a component of
   * the shapeset; false otherwise
   */

    public boolean remove (MoebiusShape aShape) {
        return remove(aShape, true);
    }

  public boolean remove (MoebiusShape aShape, boolean doFirePropertyChange) {
    if( shapes.removeElement( aShape ) ) {
      
      try {
	aShape.getClass().getMethod( "removePropertyChangeListener", new Class[] { PropertyChangeListener.class } )
	  .invoke( aShape, new Object[] { this } );
      }
      catch( Exception e ) { }
      
      if (doFirePropertyChange) {
	firePropertyChange( "removed shape", aShape, null );
      }

      return true;

    } else return false;
  } 

  public void remove (MoebiusShape[] shape) {
    for (int i = 0;i < shape.length;i++) {
      remove (shape[i]);
    } 
  } 


  /**
   * Makes the shapeset empty.
   */

  public void clear () {
      removeAll();
  } 
  
  public void removeAll() {
      Vector tmpShapes = (Vector)shapes.clone();
      
      for( int i=0; i<tmpShapes.size(); i++ ) {
	  remove( (MoebiusShape)tmpShapes.elementAt(i), false );
      }
      firePropertyChange( "removed all shapes", null, null );

  }

    public int getNumOfShapes() {
      return shapes.size();
    }


  public void draw (MoebiusGraphics G) {
    int elementCount = shapes.size ();

    G.popShape();

    for (int i = 0;i < elementCount;i++) {
      MoebiusShape   aShape = (MoebiusShape) (shapes.elementAt (i));

      G.push (aShape);
      aShape.draw (G);
      G.popShape ();
    } 

    G.push( this );
  } 

  public Vector getTools () {
    return null;
  } 

}

