/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.viewer;

import java.awt.AWTEvent;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.image.ImageObserver;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.Serializable;

import javax.swing.JComponent;

import de.jtem.moebiusViewer.AbstractBounder;
import de.jtem.moebiusViewer.AbstractDrawer;
import de.jtem.moebiusViewer.AbstractPicker;
import de.jtem.moebiusViewer.Attributes2D;

public class AbstractJViewerCanvas extends JComponent implements Serializable, ActionListener,
	       PropertyChangeListener, ImageObserver {

    private static final long serialVersionUID = 1L;

    public AbstractViewer viewer;

    protected Dimension           size = new Dimension();

    private boolean               shouldRepaint, working;

    public AbstractJViewerCanvas( AbstractDrawer drawer,
				  AbstractPicker picker,
				  AbstractBounder bounder ) {

	viewer = new AbstractViewer( this, drawer, picker, bounder );

	viewer.setAttributes( new Attributes2D() );

	enableEvents( AWTEvent.MOUSE_EVENT_MASK );
	enableEvents( AWTEvent.KEY_EVENT_MASK );

	setPreferredSize( new Dimension( 200, 200 ) );
	setMinimumSize(   new Dimension(  50,  50 ) );

	setOpaque( true );
    }

    public boolean isFocusTraversable () {
	return true;
    }

    public void propertyChange( PropertyChangeEvent e ) {
	viewer.propertyChange( e );
    }

    private void mousePressed (MouseEvent e) {
	viewer.mousePressed( e );
    }

    private void mouseReleased (MouseEvent e) {
	viewer.mouseReleased( e );
    }

    public void actionPerformed (ActionEvent e) {
	requestFocus ();
    }

    public void doLayout () {

	getSize(size);

	viewer.setSize( size );
    }


    public void validate () {
	super.validate();

	getSize(size);

	viewer.setSize( size );

	if( viewer.couldNotEncompass ) {
	    viewer.encompass( viewer.boundWhenEncompassFailed );
	}
    }

    protected void paintComponent( Graphics rg ) {

	rg = getComponentGraphics( rg ).create();

	getSize(size);

	if( size.width <= 0 || size.height <= 0 )
	    return;
        rg.setColor(getBackground());
	rg.fillRect (0, 0, size.width, size.height);

	viewer.setSize( size );
	viewer.draw (rg);
    }

    public void repaint() {
        repaint(100l, 0, 0, Integer.MAX_VALUE, Integer.MAX_VALUE);
    }
    public void repaint(int x, int y, int w, int h) {
	repaint(100l, x, y, w, h);
    }
    public void repaint(long l) {
        repaint(l, 0, 0, Integer.MAX_VALUE, Integer.MAX_VALUE);
    }

    public void repaint(long l, int x, int y, int w, int h) {
	if(working) shouldRepaint=true;
	else super.repaint(l, x, y, w, h);
    }

    protected void processMouseEvent(MouseEvent e ) {
      try {
        working=true;
	switch(e.getID())
	{
	  case MouseEvent.MOUSE_PRESSED:
            mousePressed(e);
	    break;
	  case MouseEvent.MOUSE_RELEASED:
	    mouseReleased(e);
	    break;
	  case MouseEvent.MOUSE_ENTERED:
	  case MouseEvent.MOUSE_EXITED:
	    break;
	  default:
	    requestFocus();
	}
        super.processMouseEvent(e);
        if(shouldRepaint)
	{
	    getSize( size );
	    paintImmediately( 0, 0, size.width, size.height );
	}
        shouldRepaint=false;
      }
      finally {
        working=false;
      }
    }

    protected void processMouseMotionEvent(MouseEvent e) {
      try {
        working=true;
        super.processMouseMotionEvent(e);
        if(shouldRepaint)
	{
	    getSize( size );
	    paintImmediately( 0, 0, size.width, size.height );
	}
	shouldRepaint=false;
      }
      finally {
        working=false;
      }
    }


    protected void processTypedKey( char typedKey ) {
    }

    protected void processKeyEvent( KeyEvent e ) {

	try {
	    working=true;

	    if( e.getID() == KeyEvent.KEY_TYPED ) {
		processTypedKey( e.getKeyChar() );
	    } else {
		super.processKeyEvent( e );
	    }

	    if( shouldRepaint )
		{
		    getSize( size );
		    paintImmediately( 0, 0, size.width, size.height );
		}

	    shouldRepaint=false;
	}
	finally {
	    working=false;
	}
    }
}

