/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2009, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.moebiusViewer.viewer;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.PopupMenu;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.image.BufferedImage;
import java.awt.image.Raster;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.Hashtable;
import java.util.Vector;

import de.jtem.moebiusViewer.AbstractBounder;
import de.jtem.moebiusViewer.AbstractDrawer;
import de.jtem.moebiusViewer.AbstractPSWriter;
import de.jtem.moebiusViewer.AbstractPicker;
import de.jtem.moebiusViewer.Attributes;
import de.jtem.moebiusViewer.MoebiusGraphics;
import de.jtem.moebiusViewer.MoebiusShape;
import de.jtem.moebiusViewer.MoebiusTool;
import de.jtem.moebiusViewer.MoebiusTransform;
import de.jtem.moebiusViewer.RenderTreeInfo;
import de.jtem.moebiusViewer.shape.MoebiusTransformer;
import de.jtem.moebiusViewer.shape.Rectangle;
import de.jtem.moebiusViewer.shape.ShapeSet;
import de.jtem.moebiusViewer.util.MoebiusShapeStack;
import de.jtem.moebiusViewer.util.MoebiusTransformStack;

public class AbstractViewer
    implements Serializable, ActionListener,
	       PropertyChangeListener {

    private static final long serialVersionUID = 1L;

    MoebiusTransformStack backgroundTransformPath = new MoebiusTransformStack();

    MoebiusTool                   currentTool;
    MoebiusTool                   backgroundTool;

    Hashtable                     toolTable = new Hashtable();

    public ShapeSet                      shapeSet;
    public MoebiusTransformer            transformShape;

    AbstractDrawer                drawerContext;
    AbstractPicker                pickerContext;
    public AbstractBounder               bounderContext;

    PopupMenu                     backgroundPop = new PopupMenu ();
    PopupMenu                     toolPop       = new PopupMenu ();

    protected Dimension           size = new Dimension();

    Component viewerComponent;

    AbstractViewer( Component viewerComponent ) {

	this.viewerComponent = viewerComponent;

	toolPop.add ("no tools");
	viewerComponent.add (toolPop);
	viewerComponent.add (backgroundPop);

	viewerComponent.requestFocus();
    }

    public AbstractViewer( Component viewerComponent,
			   AbstractDrawer drawer, AbstractPicker picker, AbstractBounder bounder) {
	this ( viewerComponent);

	transformShape    = new MoebiusTransformer();
	shapeSet          = new ShapeSet();

	shapeSet.      addPropertyChangeListener( this );
	transformShape.addPropertyChangeListener( this );

	transformShape.setShape( shapeSet );

	backgroundTransformPath.push( transformShape.getTransform() );

	MoebiusShapeStack tmp = new MoebiusShapeStack();

	tmp.push( transformShape );

	selectCurrentTool( transformShape, tmp, backgroundTransformPath );

	backgroundTool = currentTool;

	deactivateTool( currentTool );

	drawerContext  = drawer;
	pickerContext  = picker;
	bounderContext = bounder;

	pickerContext. setAttributes( drawerContext.getAttributes() );
	bounderContext.setAttributes( drawerContext.getAttributes() );
    }

    public Dimension getSize() {
	return new Dimension( size );
    }

    public Dimension getSize( Dimension target ) {
	if( target == null )
	    return new Dimension( size );

	target.setSize( size );

	return target;
    }

    public void setSize( Dimension size ) {
	this.size.setSize( size );

	drawerContext.setSize( size );
	pickerContext.setSize( size );
	bounderContext.setSize( size );
    }

    public void setFont (Font aFont) {
	drawerContext.setFont (aFont);
	pickerContext.setFont (aFont);
	bounderContext.setFont (aFont);
	repaint ();
    }

    public void propertyChange( PropertyChangeEvent e ) {

      String s = e.getPropertyName();

      if( s == "removed shape" ) {
	  if( !contains( (MoebiusShape)e.getOldValue() ) ) {
	      toolTable.remove( e.getOldValue() );
	  }
	  invalidate();
      } else if( s == "added shape" ) {
	  invalidate();
      } else if( s == "invalidate" ) {
	  invalidate();
      }

      repaint();
    }

    public boolean contains( MoebiusShape shape ) {
	return RenderTreeInfo.contains( transformShape, shape );
    }

    public AbstractBounder getBounderContext () {
	return bounderContext;
    }

    public void setBounderContext( AbstractBounder v) {
	this.bounderContext = v;
    }

    public AbstractPicker getPickerContext() {
	return pickerContext;
    }

    public void setPickerContext( AbstractPicker v) {
	this.pickerContext = v;
    }

    public AbstractDrawer getDrawerContext () {
	return drawerContext;
    }

    public void setDrawerContext( AbstractDrawer v) {
	this.drawerContext = v;
    }

    public void clearShapeSet () {
	removeAll();
    }

    public void removeAll() {
	shapeSet.removeAll();
    }

    public void set( MoebiusShape[] someShapes ) {
	shapeSet.set( someShapes );
    }

    public void set( int index, MoebiusShape aShape ) {
	shapeSet.set( index, aShape );
    }

    public void add (MoebiusShape aShape) {
	shapeSet.add (aShape);
    }

    public void add (MoebiusShape[] aShape) {
	shapeSet.add (aShape);
    }

    public void remove (MoebiusShape aShape) {
	shapeSet.remove (aShape);
    }

    public void remove (MoebiusShape[] aShape) {
	shapeSet.remove (aShape);
    }

    public MoebiusShape getSelection () {
	return currentTool == null ? null : currentTool.getContext();
    }

    public MoebiusTool getCurrentTool () {
	return currentTool;
    }

    public void deactivateTool( MoebiusTool aTool ) {
	if (aTool != null) {
	    viewerComponent.removeMouseListener (aTool);
	    viewerComponent.removeMouseMotionListener (aTool);
	}
    }

    public void activateTool( MoebiusTool aTool ) {

	if( aTool == null )
	    return;

	aTool.setViewer( this );

	viewerComponent.addMouseListener (aTool);
	viewerComponent.addMouseMotionListener (aTool);

    }

    public void setAttributes( Attributes a){
	drawerContext.setAttributes(a);
	pickerContext.setAttributes(a);
	bounderContext.setAttributes(a);
    }

    public Attributes getAttributes(){
	return drawerContext.getAttributes();
    }

    public Menu getMenu () {
	return toolPop;
    }

    class BackgroundSelectAction implements ActionListener {

	MoebiusShape selectedBackground;
	MoebiusTransformStack selectedTransformPath;

	BackgroundSelectAction ( MoebiusShape aBackground, MoebiusTransformStack aTransformPath ) {
	    selectedBackground = aBackground;
	    selectedTransformPath = new MoebiusTransformStack( aTransformPath );
	}

	public void actionPerformed (ActionEvent e) {

	    backgroundTool = (MoebiusTool)toolTable.get( selectedBackground );

	    if( backgroundTool == null && selectedBackground.getTools().size() != 0 ) {
		backgroundTool = (MoebiusTool)selectedBackground.getTools().elementAt(0);
	    }

	    backgroundTransformPath
		= new MoebiusTransformStack( selectedTransformPath );

	    backgroundTool.setContext( selectedBackground );
	    backgroundTool.setViewer( AbstractViewer.this );
	    backgroundTool.setTransformPath( backgroundTransformPath );

	    repaint();
	}
    }

    class ToolSelectAction implements ActionListener {
	MoebiusTool   tool;
	MoebiusShape  contextOfTool;
	boolean       contextIsBackground;

	ToolSelectAction ( MoebiusTool tool, boolean contextIsBackground ) {
	    this.tool                = tool;
	    this.contextIsBackground = contextIsBackground;
	}

	public void actionPerformed( ActionEvent e ) {
	    toolTable.remove( tool.getContext() );
	    toolTable.put( tool.getContext(), tool );

	    if( contextIsBackground ) {
		backgroundTool = tool;

		backgroundTool.setViewer( AbstractViewer.this );
		backgroundTool.setTransformPath( backgroundTransformPath );

		repaint();
	    }

	    activateTool( tool );
	    deactivateTool( tool );
	}
    }

    void createToolPop( MoebiusShape selection, boolean pickedBackground ) {
	toolPop.removeAll ();
	Vector toolsOfSelection = selection.getTools ();
	if (toolsOfSelection != null && toolsOfSelection.size() != 0 ) {
	    boolean setCurrentTool = false;
	    for( int i = 0;i < toolsOfSelection.size (); i++ ) {
		if (toolsOfSelection.elementAt (i) != null) {
		    MoebiusTool aTool = (MoebiusTool)toolsOfSelection.elementAt(i);
		    try {
			aTool = (MoebiusTool)aTool.clone();
		    }
		    catch( CloneNotSupportedException cnse ) {
			System.out.println( "could not clone Tool" );
		    }
		    if( !setCurrentTool ) {
			//  setCurrentTool( aTool );
			setCurrentTool = true;
		    }
		    MenuItem anItem = new MenuItem (aTool.getLabel ());
		    aTool.setContext( selection );
		    anItem.addActionListener
			( new ToolSelectAction( aTool, pickedBackground ) );
		    toolPop.add (anItem);
		}
	    }
	    if( !setCurrentTool )
		toolPop.add ("no tools");
	} else toolPop.add ("no tools");
    }

    void createBackgroundPop( MoebiusShapeStack currentPickPath,
			      MoebiusTransformStack currentPickTransformPath ) {
	backgroundPop.removeAll ();
	for( int i = 0;i < currentPickPath.size () - 1;i++) {
	    if( currentPickPath.elementAt (i) != null) {

		MoebiusShape aBackground = (MoebiusShape)currentPickPath.elementAt(i);
		MenuItem  anItem =
		    new MenuItem( aBackground.getClass().getName() + i);
		anItem.addActionListener
		    ( new BackgroundSelectAction( aBackground, currentPickTransformPath ));
		backgroundPop.add (anItem);
	    }
	}
    }

    void selectCurrentTool( MoebiusShape currentSelection,
			    MoebiusShapeStack currentPickPath,
			    MoebiusTransformStack currentPickTransformPath) {

	currentTool = (MoebiusTool)toolTable.get( currentSelection );

	// take default tool if no tool is in the table
	if (currentTool == null) {

	    Vector toolvec = currentSelection.getTools();
	    if ( toolvec != null && toolvec.size() > 0 ) {
		currentTool = (MoebiusTool)toolvec.elementAt(0);
		try {
		    currentTool = (MoebiusTool)currentTool.clone();
		}
		catch( CloneNotSupportedException cnse ) {
		    System.out.println( "could not clone Tool" );
		}
	    }
	}
	if( currentTool != null ) {
	    currentTool.setContext( currentSelection );
	    currentTool.setTransformPath( currentPickTransformPath );

	    activateTool( currentTool );
	}
    }

     void mousePressed (MouseEvent e) {
	MoebiusShapeStack currentPickPath;
	MoebiusTransformStack currentPickTransformPath;
	MoebiusShape currentSelection;

	boolean pickedBackground = !pick( e );

	if( !pickedBackground ) {
	    currentPickPath          = pickerContext.getPickPath();
	    currentPickTransformPath = pickerContext.getPickTransformPath();
	    currentSelection         = currentPickPath.peek();
	} else {
	    currentPickPath          = null;
	    currentPickTransformPath = backgroundTransformPath;
	    if( backgroundTool != null )
		currentSelection = backgroundTool.getContext();
	    else
		currentSelection = null;
	}

	if( e.isControlDown() ) {
	    if( currentSelection != null ) {
		createToolPop( currentSelection, pickedBackground );
		toolPop.show( viewerComponent, e.getX (), e.getY() );
	    }

	} else if( e.isShiftDown() ) {
	    if( !pickedBackground ) {
		createBackgroundPop( currentPickPath, currentPickTransformPath );
		backgroundPop.show ( viewerComponent, e.getX (), e.getY ());
	    }
	} else {
	    if( currentSelection != null ) {

		selectCurrentTool( currentSelection, currentPickPath, currentPickTransformPath );

		if( currentTool != null ) {
		    currentTool.mousePressed (e);
		}
	    }
	}

	viewerComponent.requestFocus ();
    }

     void mouseReleased (MouseEvent e) {

	deactivateTool( currentTool );
	currentTool = null;

	viewerComponent.requestFocus ();
    }

    public void actionPerformed (ActionEvent e) {
	viewerComponent.requestFocus ();
    }

    public void invalidate () {
	viewerComponent.invalidate ();
    }

    public void repaint() {
        viewerComponent.repaint();
    }

    public void paintOffScreen( BufferedImage anImage ) {

	Graphics anImageGraphics = anImage.getGraphics();

	anImageGraphics.setColor (viewerComponent.getBackground());
	anImageGraphics.fillRect (0, 0, anImage.getWidth(), anImage.getHeight());

	draw (anImageGraphics);
    }


    public void writePPM( String filename ) throws IOException {
	writePPM( new File( filename ) );
    }

    public void writePPM( File file ) throws IOException {

        viewerComponent.getSize(size);
	BufferedImage image = new BufferedImage( size.width, size.height, BufferedImage.TYPE_INT_RGB );

	paintOffScreen( image );

	Raster raster = image.getData();

	int [] intData = raster.getPixels( 0, 0, size.width, size.height, (int [])null );

	byte [] byteData = new byte[ intData.length];

	for( int i=0; i<intData.length; i++ ) {
	    byteData[i] = (byte)intData[i];
	}

	FileOutputStream output = new FileOutputStream( file );

	String header = "P6\n# created by de.jtem.moebiusViewer\n" + size.width + " " + size.height + "\n255\n";

	output.write( header.getBytes() );
	output.write( byteData );
	output.close();
    }

    static String directory = null;

    public void writePPM() {
	Component   component = viewerComponent;

	while (component.getParent () != null) {
	    component = component.getParent ();
	}

	try {

	    FileDialog choose = new FileDialog ((Frame) component, "ChooseFile", FileDialog.SAVE);

                        choose.setDirectory( directory );

	    choose.show ();

	    String fileName = choose.getFile();

	    if( fileName == null )
		return;

	    File file = new File( directory = choose.getDirectory(), fileName );

	    writePPM( file );

	}
	catch ( IOException e ) {}
	catch (ClassCastException e) {}

    }

    public void writePS( MoebiusShape shape, Rectangle aBound ) {
	Component component = viewerComponent;

	while (component.getParent () != null) {
	    component = component.getParent ();
	}

	try {
	    FileDialog        choose = new FileDialog ((Frame) component,
						       "ChooseFile", FileDialog.SAVE);

            choose.setDirectory( directory );

	    choose.show ();

	    String fileName = choose.getFile();

	    if( fileName == null )
		return;

	    File file = new File( directory = choose.getDirectory(), fileName );

	    AbstractPSWriter psWriter = getDrawerContext().getPSWriter();

	    psWriter.setBound( aBound );
	    psWriter.setShape( shape );

	    psWriter.write( file );
	}
	catch (ClassCastException e) {}

    }

    public void writePS( Rectangle aBound ) {
	writePS( transformShape, aBound );
    }

    public void writePS( Dimension d ) {
	writePS( transformShape, new Rectangle( 0, 0, d.width, d.height ) );
    }

    public void writePS() {
        viewerComponent.getSize(size);
	writePS( size );
    }

    public void draw( MoebiusGraphics g ) {
	transformShape.draw( g );
    }

    public void draw (Graphics G) {

	drawerContext.draw (G, transformShape);
	drawerContext.clear();

	if( backgroundTool != null ) {
	    drawerContext.push( backgroundTool.getContextTransform() );
	    drawerContext.draw (G, backgroundTool);
	    drawerContext.popTransform();
	}

	// the current tool is only rendered when it mouse is pressed.
	// than the transform path is set correctly
	if( currentTool != backgroundTool && currentTool != null ) {
	    drawerContext.push( currentTool.getContextTransform() );
	    drawerContext.draw (G, currentTool);
	    drawerContext.popTransform();

	}
    }

    public boolean pick (MouseEvent e) {

	boolean isPicked = pickerContext.pick (e, transformShape );

	return isPicked;
    }

    public de.jtem.moebiusViewer.shape.Rectangle getBound () {
	return bounderContext.getBound( transformShape );
    }

    public void encompass() {
	encompass( false );
    }
    public void encompass( boolean reset ) {
      de.jtem.moebiusViewer.shape.Rectangle aBound = this.getBound();

      if( aBound != null )
	encompass( aBound, reset );
    }

    boolean couldNotEncompass = false;

    de.jtem.moebiusViewer.shape.Rectangle boundWhenEncompassFailed;

    public void encompass( de.jtem.moebiusViewer.shape.Rectangle aBound ) {
	encompass( aBound, false );
    }
    public void encompass( de.jtem.moebiusViewer.shape.Rectangle aBound, boolean reset ) {

	if( aBound == null )
	  throw new IllegalArgumentException( "bound must not be null" );

	viewerComponent.getSize (size);

	double              width = size.width;
	double              height = size.height;

	if( width == 0 || height == 0 ) {
	  couldNotEncompass = true;
	  boundWhenEncompassFailed = aBound;
	  return;
	}

	couldNotEncompass = false;

	double              scale = 1;
	MoebiusTransform    transform = transformShape.getTransform ();

	if( reset )
	    transform.assignIdentity();

	if (aBound.w != 0 || aBound.h != 0) {
	  scale = aBound.w / width < aBound.h / height ? height / aBound.h
	    : width / aBound.w;
	}

	scale *= 0.9;

	double dx = width/2.0 - (aBound.x +  aBound.w/2.0);
	double dy = height/2.0 - (aBound.y + aBound.h/2.0);

	transform.scale (scale);
	transform.translate(dx, dy);
	center ();
    }

    public void center () {
	de.jtem.moebiusViewer.shape.Rectangle aBound = getBound ();
	MoebiusTransform transform = transformShape.getTransform ();

	if (aBound != null) {
	    viewerComponent.getSize (size);
	    transform.translate (((double) size.width) / 2.0
				 - (aBound.x + aBound.w / 2.0),
				 ((double) size.height) / 2.0
				 - (aBound.y + aBound.h / 2.0));
	}

    }

    public void invertInUnitCircle () {

	MoebiusTransform transform = transformShape.getTransform ();

	transform.assign(transform.getB(), transform.getA(), transform.getD(), transform.getC());
    }

}

